/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <ansilove.h>
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */
#include <sys/mman.h>                   /* MAP_FAILED */

#include "sauce.h"

#define ENABLE_ANSI
#define ENABLE_ARTWORX
#define ENABLE_BINARY
#define ENABLE_ICEDRAW
#define ENABLE_PCBOARD
#define ENABLE_TUNDRA
#define ENABLE_XBIN

/*

* ansilove_ansi()
* ansilove_artworx()
* ansilove_binary()
* ansilove_icedraw()
* ansilove_pcboard()
* ansilove_tundra()
* ansilove_xbin()

*/

typedef int(*decode_func_t)(struct ansilove_ctx*, struct ansilove_options*);

struct _abydos_plugin_handle_t{
    decode_func_t decode_func;
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_al_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_ANSI
        "text/x-ansi",
#endif
#ifdef ENABLE_ARTWORX
        "image/x-artworx",
#endif
#ifdef ENABLE_BINARY
        "text/x-binary",
#endif
#ifdef ENABLE_ICEDRAW
        "image/x-icedraw",
#endif
#ifdef ENABLE_PCBOARD
        "text/x-pcboard",
#endif
#ifdef ENABLE_TUNDRA
        "text/x-tundra",
#endif
#ifdef ENABLE_XBIN
        "image/x-xbin",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_al_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_ARTWORX
    if(!strcasecmp(mime_type,"image/x-artworx")){
        h->decode_func = ansilove_artworx;
    }else
#endif
#ifdef ENABLE_BINARY
    if(!strcasecmp(mime_type,"text/x-binary")){
        h->decode_func = ansilove_binary;
    }else
#endif
#ifdef ENABLE_ICEDRAW
    if(!strcasecmp(mime_type,"image/x-icedraw")){
        h->decode_func = ansilove_icedraw;
    }else
#endif
#ifdef ENABLE_PCBOARD
    if(!strcasecmp(mime_type,"text/x-pcboard")){
        h->decode_func = ansilove_pcboard;
    }else
#endif
#ifdef ENABLE_TUNDRA
    if(!strcasecmp(mime_type,"text/x-tundra")){
        h->decode_func = ansilove_tundra;
    }else
#endif
#ifdef ENABLE_XBIN
    if(!strcasecmp(mime_type,"image/x-xbin")){
        h->decode_func = ansilove_xbin;
    }else
#endif
    {
        h->decode_func = ansilove_ansi;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_al_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_al_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    sauce_t sauce;
    struct ansilove_options opts;
    struct ansilove_ctx ctx;
    int res;

    ansilove_init(&ctx,&opts);

#ifdef ENABLE_ICEDRAW
    if(ansilove_icedraw == h->decode_func){
        opts.icecolors = 1;
        opts.bits = 8;
        h->info->pixel_ratio = 20.0/24;
    }else
#endif
#ifdef ENABLE_XBIN
    if(ansilove_xbin == h->decode_func){
        opts.bits = 8;
        h->info->pixel_ratio = 20.0/24;
    }else
#endif
    {
        /* handle SAUCE */
        if(sauce_validate(data,len,&sauce)){
            opts.icecolors = sauce_ANSiFlags_iCE_Color(&sauce);
            opts.bits = sauce_ANSiFlags_Letter_Spacing(&sauce);
            if(!sauce_ANSiFlags_Aspect_Ratio(&sauce))
                h->info->pixel_ratio = 20.0/(3*opts.bits);
            len -= sauce.total_size;
#ifdef ENABLE_BINARY
            if(ansilove_binary == h->decode_func)
                opts.columns = sauce_BinaryText_width(&sauce);
#endif
        }else{
            opts.bits = 9;
            h->info->pixel_ratio = 20.0/27;
        }
    }

    ctx.buffer = (uint8_t*)data;
    ctx.length = len;

    res = h->decode_func(&ctx,&opts);
    DEBUG_p("decode_func() result: %d, %d",res,ctx.error);

    if(0==res){
        abydos_t *ar = abydos_create("image/png");
        abydos_from_data(ar,(const char*)ctx.png.buffer,ctx.png.length);
        h->surface = abydos_get_image_surface(ar);
        abydos_get_size(ar,&h->info->width,&h->info->height);
        abydos_destroy(ar);
    }

    ctx.buffer = MAP_FAILED;
    ansilove_clean(&ctx);
    return res;
}


static cairo_surface_t *
_al_get_image_surface(abydos_plugin_handle_t *h,
                      int page,
                      cairo_bool_t *layer_visibility,
                      int variant,
                      double time,
                      int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _al_get_mime_types,
        .new = _al_new,
        .free = _al_free,
        .create_from_data = _al_create_from_data,
        .get_image_surface = _al_get_image_surface,
    };
    return &p;
}
