/*
 * Written in 2021 by Magnus Bergman <magnus@snisurset.net>
 *
 * To the extent possible under law, the author(s) have dedicated all
 * copyright and related and neighboring rights to this software to the
 * public domain worldwide.
 *
 * This software is distributed without any warranty.
 *
 * See: http://creativecommons.org/publicdomain/zero/1.0/
 */
#include <debug.h>
#include <endian.h>
#include <fcntl.h>                      /* open() */
#include <stddef.h>                     /* NULL */
#include <stdio.h>                      /* printf() */
#include <string.h>
#include <sys/mman.h>                   /* mmap() */
#include <sys/stat.h>
#include <unistd.h>                     /* close() */

#include "cgmparse.h"
#include "cgmread-binary.h"
#include "cgmread-text.h"


static int
_mmap_open(const char *filename, int *r_fd,const void **r_data, size_t *r_len)
{
    struct stat statbuf;

    if(stat(filename,&statbuf))
        return -1;

    *r_fd = open(filename,O_RDONLY);
    if(*r_fd<0)
        return -1;
    *r_data = mmap(NULL,statbuf.st_size,PROT_READ,MAP_SHARED|MAP_POPULATE,*r_fd,0);
    if(MAP_FAILED==*r_data){
        close(*r_fd);
        return -1;
    }
    *r_len = statbuf.st_size;
    return 0;
}


static void
_mmap_close(int fd,const void *data,size_t len)
{
    munmap((void*)data,len);
    close(fd);
}


static int
_on_entry(void *userdata, const cgm_element_t *e, cgm_read_t *r)
{
    const char *name = "";
    switch(e->klass){
    case 4:
        switch(e->id){
        case 0x09: name = "Cell Array";break;
        }
        break;
    }

    printf("Class: %02d, ID: %02x, %s\n", e->klass, e->id, name);
    return 0;
}


static int
_cgm_parse(const void *data, size_t len){
    if(len<4)
        return -1;
    if(1 == be16toh(*(uint16_t*)data) >> 5){
        DEBUG_p("binary file");
        return cgm_read_binary(data, len, _on_entry, NULL);
    }else if(!strncmp(data, "BegM", 4)){
        DEBUG_p("text file");
        return cgm_read_text(data, len, _on_entry, NULL);
    }else{
        DEBUG_p("incorrect file");
        return -1;
    }
}


int main(int argc, char **argv){
    int fd;
    const void *data;
    size_t len;

    if(argc < 2){
        printf("Usage: %s <filename.cgm>", argv[0]);
        return 1;
    }

    if(_mmap_open(argv[1], &fd, &data, &len)){
        printf("Failed to read %s", argv[1]);
        return 1;
    }

    _cgm_parse(data, len);

    _mmap_close(fd, data, len);
}
