/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <grok.h>
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static void
_grok_done()
{
    grk_deinitialize();
}


static const char **
_grok_get_mime_types()
{
    static const char *mime_types[] = {
        "image/jpeg2000",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_grok_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    return h;
}


static void
_grok_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static void
_on_info()
{
    DEBUG_p("begin");
}


static void
_on_warning()
{
    DEBUG_p("begin");
}


static void
_on_error()
{
    DEBUG_p("begin");
}


typedef void(*blit_func_t)(uint32_t*, int32_t**, int);


static void
_blit_g8(uint32_t *dst, int32_t *src[4], int count)
{
    while(count-->0){
        int g = *src[0]++;
        *dst++ = g << 16 | g << 8 | g << 0;
    }
}


static void
_blit_g16(uint32_t *dst, int32_t *src[4], int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float g = (float)*src[0]++ / INT16_MAX;
        *dstf++ = g;
    }
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
    while(count-->0){
        int g = *src[0]++ >> 6;
        *dst++ = g << 20 | g << 10 | g << 0;
    }
#else
    while(count-->0){
        int g = *src[0]++ >> 8;
        *dst++ = g << 16 | g << 8 | g << 0;
    }
#endif
}


static void
_blit_ga8(uint32_t *dst, int32_t *src[4], int count)
{
    while(count-->0){
        int a = *src[1]++;
        int g = a * *src[0]++ / 255;
        *dst++ = g << 16 | g << 8 | g << 0 | a << 24;
    }
}


static void
_blit_ga16(uint32_t *dst, int32_t *src[4], int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float a = (float)*src[1]++ / UINT16_MAX;
        float g = a * *src[0]++ / UINT16_MAX;
        *dstf++ = g;
        *dstf++ = a;
    }
#else
    while(count-->0){
        int a = *src[1]++ >> 1;
        int g = (a * *src[0]++ / UINT16_MAX) >> 7;
        *dst++ = g << 16 | g << 8 | g << 0 | (a>>7) << 24;
    }
#endif
}


static void
_blit_rgb8(uint32_t *dst, int32_t *src[4], int count)
{
    while(count-->0){
        int r = *src[0]++;
        int g = *src[1]++;
        int b = *src[2]++;
        *dst++ = r << 16 | g << 8 | b << 0;
    }
}


static void
_blit_rgb16(uint32_t *dst, int32_t *src[4], int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float r = *src[0]++ / UINT16_MAX;
        float g = *src[1]++ / UINT16_MAX;
        float b = *src[2]++ / UINT16_MAX;
        *dstf++ = r;
        *dstf++ = g;
        *dstf++ = b;
    }
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
    while(count-->0){
        int r = *src[0]++ >> 6;
        int g = *src[1]++ >> 6;
        int b = *src[2]++ >> 6;
        *dst++ = r << 20 | g << 10 | b << 0;
    }
#else
    while(count-->0){
        int r = *src[0]++ >> 8;
        int g = *src[1]++ >> 8;
        int b = *src[2]++ >> 8;
        *dst++ = r << 16 | g << 8 | b << 0;
    }
#endif
}


static void
_blit_rgba8(uint32_t *dst, int32_t *src[4], int count)
{
    while(count-->0){
        int a = *src[3]++;
        int r = a * *src[0]++ / 255;
        int g = a * *src[1]++ / 255;
        int b = a * *src[2]++ / 255;
        *dst++ = r << 16 | g << 8 | b << 0 | a << 24;
    }
}


static void
_blit_rgba16(uint32_t *dst, int32_t *src[4], int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float a = (float)*src[3]++ / UINT16_MAX;
        float r = a * *src[0]++ / UINT16_MAX;
        float g = a * *src[1]++ / UINT16_MAX;
        float b = a * *src[2]++ / UINT16_MAX;
        *dstf++ = r;
        *dstf++ = g;
        *dstf++ = b;
        *dstf++ = a;
    }
#else
    while(count-->0){
        int a = *src[3]++ >> 1;
        int r = (a * *src[0]++ / UINT16_MAX) >> 7;
        int g = (a * *src[1]++ / UINT16_MAX) >> 7;
        int b = (a * *src[2]++ / UINT16_MAX) >> 7;
        *dst++ = r << 16 | g << 8 | b << 0 | (a>>7) << 24;
    }
#endif
}


static cairo_surface_t *
_surface_from_image(grk_image *img)
{
    int w = img->x1 - img->x0;
    int h = img->y1 - img->y0;
    cairo_format_t format;
    cairo_surface_t *surface;
    blit_func_t blit;
    uint32_t *dst;
    int dst_stride;
    int32_t *src[4];
    int src_skip[4];
    int i;
    if(img->numcomps <= 0 || img->numcomps > 4)
        return NULL;
    for(i=0; i<img->numcomps; ++i){
        grk_image_comp *comp = img->comps + i;
        if(comp->w != w || comp->h != h)
            return NULL;
        src[i] = comp->data;
        if(!src[i])
            return NULL;
        src_skip[i] = comp->stride - w;
    }
    switch(img->numcomps){
    case 1: /* gray scale */
        format = CAIRO_FORMAT_RGB24;
        if(8 == img->comps[0].prec){
            blit = _blit_g8;
        }else if(16 == img->comps[0].prec){
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            format = CAIRO_FORMAT_RGB96F;
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
            format = CAIRO_FORMAT_RGB30;
#endif
            blit = _blit_g16;
        }else{
            return NULL;
        }
        break;
    case 2: /* gray scale + alpha */
        format = CAIRO_FORMAT_ARGB32;
        if(8 == img->comps[0].prec &&
           8 == img->comps[1].prec)
        {
            blit = _blit_ga8;
        }else if(16 == img->comps[0].prec &&
                 16 == img->comps[1].prec)
        {
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            format = CAIRO_FORMAT_RGBA128F;
#endif
            blit = _blit_ga16;
        }else{
            return NULL;
        }
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, w, h);
        break;
    case 3: /* RGB */
        format = CAIRO_FORMAT_RGB24;
        if(8 == img->comps[0].prec &&
           8 == img->comps[1].prec &&
           8 == img->comps[2].prec)
        {
            blit = _blit_rgb8;
        }else if(16 == img->comps[0].prec &&
                 16 == img->comps[1].prec &&
                 16 == img->comps[2].prec)
        {
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            format = CAIRO_FORMAT_RGB96F;
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
            format = CAIRO_FORMAT_RGB30;
#endif
            blit = _blit_rgb16;
        }else{
            return NULL;
        }
        break;
    case 4: /* RGB + alpha */
        format = CAIRO_FORMAT_ARGB32;
        if(8 == img->comps[0].prec &&
           8 == img->comps[1].prec &&
           8 == img->comps[2].prec &&
           8 == img->comps[3].prec)
        {
            blit = _blit_rgba8;
        }else if(16 == img->comps[0].prec &&
                 16 == img->comps[1].prec &&
                 16 == img->comps[2].prec &&
                 16 == img->comps[3].prec)
        {
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(2,7,2)
            format = CAIRO_FORMAT_RGBA128F;
#endif
            blit = _blit_rgba16;
        }else{
            return NULL;
        }
        break;
    }
    surface = cairo_image_surface_create(format, w, h);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    dst_stride = cairo_image_surface_get_stride(surface) / sizeof(uint32_t);
    while(h-->0){
        blit(dst, src, w);
        dst += dst_stride;
        for(i=0; i<img->numcomps; ++i)
            src[i] += src_skip[i];
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_grok_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    //uint16_t numTiles = 0;
    grk_decompress_parameters param;
    grk_stream_params stream_params;
    grk_codec *codec = NULL;
    grk_header_info header;
    grk_image *image;
    int res = -1;

    memset(&param, 0, sizeof(grk_decompress_parameters));
    param.compressionLevel = GRK_DECOMPRESS_COMPRESSION_LEVEL_DEFAULT;
    grk_decompress_set_default_params(&param.core);
#ifdef USE_DEBUG
    param.verbose_ = true;
#endif

    memset(&stream_params,0,sizeof(stream_params));
    stream_params.buf = (uint8_t*)data;
    stream_params.len = len;

    codec = grk_decompress_init(&stream_params, &param.core);
    if(!codec){
        DEBUG_p("failed to set up decompressor");
        goto out;
    }

    memset(&header, 0, sizeof(grk_header_info));
    if(!grk_decompress_read_header(codec, &header)){
        DEBUG_p("failed to read the header");
        goto out;
    }

    if(!grk_decompress(codec, NULL))
        goto out;

    image = grk_decompress_get_composited_image(codec);
    if(!image){
        DEBUG_p("failed to retrieve image");
        goto out;
    }

    h->info->width = image->x1 - image->x0;
    h->info->height = image->y1 - image->y0;
    h->surface = _surface_from_image(image);
    res = h->surface ? 0 : -1;
out:
    if(codec)
        grk_object_unref(codec);
    return res;
}


static cairo_surface_t *
_grok_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = _grok_done,
        .get_mime_types = _grok_get_mime_types,
        .new = _grok_new,
        .free = _grok_free,
        .create_from_data = _grok_create_from_data,
        .get_image_surface = _grok_get_image_surface
    };
    grk_initialize(NULL, 0);
    grk_set_msg_handlers(_on_info, NULL, _on_warning, NULL, _on_error, NULL);
    return &p;
}
