/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <math.h>                       /* ldexp() */
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


#define get8(a) (*(uint8_t*)(a))
#define get16(a) le16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    double width;
    double height;
    cairo_surface_t *surface;
    double scale;
};


static const char **
_dcd_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-draftchoice-drawing",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_dcd_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_dcd_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static double
_get_double(const char *data)
{
    int exp = get8(data)- 0x9c;
    uint64_t tmp;
    int sign;
    double res;
    res = get8(data+1);
    tmp = get32(data+2);
    sign = tmp & (1l<<31) ? 1 : 0;
    tmp |= (1l<<31);
    res += tmp << 8;
    //DEBUG_p("%f %d %ld", res, exp, 1l<<exp);
    res = ldexp(res, exp);
    //res /= 656.221162;
    return sign ? -res : res;
}


/*
---empty.dch----------------------------------------------------------------
___pascal string_____________________________________________________
0x0d 0x44 0x52 0x41 0x46 0x54 0x20 0x43 0x48 0x4f 0x49 0x43 0x45 0x1a

0x00	0x00
___=211_____    ___=30______    ___=78______
0xd3	0x00	0x1e	0x00	0x4e	0x00	0x00	0x00
                                                __float_____
0x00	0x04	0x00	0x00	0x01	0x00	0x75	0x93
____________________________
0x02	0x86	0x3e	0x03	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
                                ___width____________________
0x00	0x00	0x04	0x00	0x82	0xd3	0x2e	0x6e
____________
0x12	0x43	0x00	0x00	0x00	0x00	0x00	0x00
                                                ___height___
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
____________________________    ___width____________________
0xa3	0xd2	0x4d	0x12	0x82	0xd3	0x2e	0x6e
____________
0x12	0xc3	0x00	0x00	0x00	0x00	0x00	0x00
                                                ___height___
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
____________________________    ___end______________________
0xa3	0xd2	0x4d	0x92	0x01	0x00	0x00	0x00

82 d3 2e 6e 12 43 = 10
82 1e a3 d2 4d 12 = 7.5
---empty2.dch---------------------------------------------------------------
___pascal string____________________________________________
0x0d	0x44	0x52	0x41	0x46	0x54	0x20	0x43
____________________________________________
0x48	0x4f	0x49	0x43	0x45	0x1a	0x00	0x00
___=211_____    ___=30______    ___=78______
0xd3	0x00	0x1e	0x00	0x4e	0x00	0x00	0x00
                                                ____________
0x00	0x04	0x00	0x00	0x01	0x00	0x74	0xd3
____________________________
0x9b	0xf9	0x53	0x56	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
                                        ____________________
0x00	0x00	0x04	0x00	0x82	0x1e	0xa3	0xd2
____________
0x4d	0x12	0x00	0x00	0x00	0x00	0x00	0x00
                                                        ____
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0xd3
____________________________            ____________________
0x2e	0x6e	0x12	0x43	0x82	0x1e	0xa3	0xd2
____________
0x4d	0x92	0x00	0x00	0x00	0x00	0x00	0x00
                                                        ____
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0xd3
____________________________    ___end______________________
0x2e	0x6e	0x12	0xc3	0x01	0x00	0x00	0x00
---empty3.dch---------------------------------------------------------------
___pascal string____________________________________________
0x0d	0x44	0x52	0x41	0x46	0x54	0x20	0x43
____________________________________________
0x48	0x4f	0x49	0x43	0x45	0x1a	0x00	0x00
__=211______    ___=30______    ___=78______
0xd3	0x00	0x1e	0x00	0x4e	0x00	0x00	0x00
                                                ____________
0x00	0x04	0x00	0x00	0x01	0x00	0x74	0xd3
____________________________
0x9b	0xf9	0x53	0x56	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
                                ___width____________________
0x00	0x00	0x04	0x00	0x82	0x1e	0xa3	0xd2
____________
0x4d	0x12	0x00	0x00	0x00	0x00	0x00	0x00
                                                ___height___
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0xd3
____________________________    ___width____________________
0x2e	0x6e	0x12	0x43	0x82	0x1e	0xa3	0xd2
____________
0x4d	0x92	0x00	0x00	0x00	0x00	0x00	0x00
                                                ___height___
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0xd3
____________________________    ___end______________________
0x2e	0x6e	0x12	0xc3	0x01	0x00	0x00	0x00
---line.dch-----------------------------------------------------------------
__pascal string_____________________________________________
0x0d	0x44	0x52	0x41	0x46	0x54	0x20	0x43
____________________________________________
0x48	0x4f	0x49	0x43	0x45	0x1a	0x00	0x00
___=211_____    ___=30______    ___=78______
0xd3	0x00	0x1e	0x00	0x4e	0x00	0x00	0x00
0x00	0x04	0x00	0x00	0x01	0x00	0x75	0x93
0x02	0x86	0x3e	0x03	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x04	0x00	0x82	0xd3	0x2e	0x6e
0x12	0x43	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
0xa3	0xd2	0x4d	0x12	0x82	0xd3	0x2e	0x6e
0x12	0xc3	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
                                ___line_____________________
0xa3	0xd2	0x4d	0x92	0x11	0x00	0x26	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x01	0x00
___________________________________________     ____________
0x7c	0x0e	0x23	0x3f	0xda	0x1b	0x7e	0x93
____________________________            ____________________
0x02	0x86	0x3e	0x03	0x82	0xb3	0x8d	0x13
____________
0xa7	0x09	0x80	0xac	0x3e	0xf7	0x5e	0x08
                                                ___end______
0x80	0xf4	0xe6	0xcd	0xb9	0x96	0x01	0x00
____________
0x00	0x00
---line2.dhc----------------------------------------------------------------
0x0d	0x44	0x52	0x41	0x46	0x54	0x20	0x43
0x48	0x4f	0x49	0x43	0x45	0x1a	0x00	0x00
0xd3	0x00	0x1e	0x00	0x4e	0x00	0x00	0x00
0x00	0x04	0x00	0x00	0x01	0x00	0x75	0x93
0x02	0x86	0x3e	0x03	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x04	0x00	0x82	0xd3	0x2e	0x6e
0x12	0x43	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
0xa3	0xd2	0x4d	0x12	0x82	0xd3	0x2e	0x6e
0x12	0xc3	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
                                ___line_____________________
0xa3	0xd2	0x4d	0x92	0x11	0x00	0x26	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x01	0x00
                                                ____________
0x7c	0x0e	0x23	0x3f	0xda	0x1b	0x7e	0xe4
____________________________            ____________________
0xb6	0xd9	0x9f	0x12	0x82	0xba	0xdc	0x2f
____________
0xef	0x0a	0x80	0xac	0x3e	0xf7	0x5e	0x08
                                                ___end______
0x80	0xf4	0xe6	0xcd	0xb9	0x96	0x01	0x00
____________
0x00	0x00
---box.dhc------------------------------------------------------------------
___pascal string____________________________________________
0x0d	0x44	0x52	0x41	0x46	0x54	0x20	0x43
____________________________________________
0x48	0x4f	0x49	0x43	0x45	0x1a	0x00	0x00
___=211_____    ___=30______    ___=78______
0xd3	0x00	0x1e	0x00	0x4e	0x00	0x00	0x00
0x00	0x04	0x00	0x00	0x01	0x00	0x75	0x93
0x02	0x86	0x3e	0x03	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x04	0x00	0x82	0xd3	0x2e	0x6e
0x12	0x43	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
0xa3	0xd2	0x4d	0x12	0x82	0xd3	0x2e	0x6e
0x12	0xc3	0x00	0x00	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x82	0x1e
                                ___box______________________
0xa3	0xd2	0x4d	0x92	0x1e	0x00	0x4e	0x00
0x00	0x00	0x00	0x04	0x00	0x00	0x01	0x00
0x7c	0x0e	0x23	0x3f	0xda	0x1b	0x7d	0xe5
0xc2	0x56	0xa6	0x13	0x81	0x34	0xd0	0x16
0x4c	0x56	0x00	0x00	0x04	0x00	0x80	0xda
0xaa	0x5c	0x99	0x11	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
0x80	0x1c	0x1a	0x86	0xbf	0x09	0x80	0xda
0xaa	0x5c	0x99	0x91	0x00	0x00	0x00	0x00
0x00	0x00	0x00	0x00	0x00	0x00	0x00	0x00
                                                ___end______
0x80	0x1c	0x1a	0x86	0xbf	0x89	0x01	0x00
____________
0x00	0x00

---4lines.dch---------------------------------------------------------------

__pascal string______________________________________________________
0x0D 0x44 0x52 0x41 0x46 0x54 0x20 0x43 0x48 0x4F 0x49 0x43 0x45 0x1A

__=0_____ __=211___ __=30____ __=78____
0x00 0x00 0xD3 0x00 0x1E 0x00 0x4E 0x00
__=0_____ __=1024__ __=0_____ __=1_____
0x00 0x00 0x00 0x04 0x00 0x00 0x01 0x00
__float_@16_(aspect ratio)___
0x7E 0x37 0x83 0x27 0x0E 0x24

0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00
          __=4_____
0x00 0x00 0x04 0x00
__width_@38__________________
0x82 0xD3 0x2E 0x6E 0x12 0x43

0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00
__height_@56_________________
0x82 0x1E 0xA3 0xD2 0x4D 0x12
__float_@62__________________
0x82 0xD3 0x2E 0x6E 0x12 0xC3

0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00
__float_@80__________________
0x82 0x1E 0xA3 0xD2 0x4D 0x92


__line___ __size___
0x11 0x00 0x26 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x01 0x00
0x7C 0x0E 0x23 0x3F 0xDA 0x1B
0x7C 0x8A 0x43 0xF8 0x75 0x34
0x82 0xD2 0xD5 0x01 0xCE 0x0F
0x7D 0x38 0x83 0x27 0x0E 0x24
0x7D 0x40 0xC4 0x99 0x45 0xD5
__line___ __size___
0x11 0x00 0x26 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x01 0x00
0x7C 0x0E 0x23 0x3F 0xDA 0x1B
0x7D 0x89 0x43 0xF8 0x75 0x34
0x82 0xD2 0xD5 0x01 0xCE 0x0F
0x7D 0x0F 0x23 0x3F 0xDA 0x1B
0x7D 0x00 0x64 0xB1 0x11 0xCD
__line___ __size___
0x11 0x00 0x26 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x01 0x00
0x7C 0x0E 0x23 0x3F 0xDA 0x1B
0x7E 0xA8 0x32 0x7A 0x58 0x07
0x82 0xD6 0x9E 0xDF 0x92 0x10
0x7D 0x60 0xE3 0x0F 0x42 0x2C
0x7D 0x80 0x84 0x6A 0xAD 0xE5
__line___ __size___
0x11 0x00 0x26 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x01 0x00
0x7C 0x0E 0x23 0x3F 0xDA 0x1B
0x7E 0xDC 0x03 0xC9 0xDD 0x44
0x82 0xD1 0x92 0x62 0x8C 0x0F
0x7D 0xE8 0xC2 0x56 0xA6 0x13
0x7D 0xE0 0x03 0xC9 0xDD 0xC4
__end____ __size___
0x01 0x00 0x00 0x00

header end: 100, header length: 86
---glass3.dch---------------------------------------------------------------
__=0________    __=200______    __=2________    __=48_______
0x00	0x00	0xc8	0x00	0x02	0x00	0x30	0x00
__float______________________
0x7f 0x77 0x95 0xbd 0xe2 0x35
__float______________________
0x7b 0x3c 0x5d 0x5f 0x9f 0xbc
0x88 0x00 0x00 0x00 0x00 0x48
0x88 0x33 0xfd 0x01 0xf1 0x45
0x7f 0xdc 0x8b 0xbe 0x0E 0x1C
0x7F 0xDC 0x8B 0xBE 0x0E 0x1C
0x7D 0xDC 0x8B 0xBE 0x0E 0x1C
0x7D 0xDC 0x8B 0xBE 0x0E 0x1C
__=3_____ __=2_____ __=1_____ __=4_____
0x03 0x00 0x02 0x00 0x01 0x00 0x04 0x00 0x28 0x00
__float______________________
0x7D 0xDC 0x8B 0xBE 0x0E 0x1C
__float______________________
0x7C 0xC7 0xDF 0x30 0xB1 0x79
__float______________________
0x81 0x9A 0x99 0x99 0x99 0x19
0x00 0x00 0x00 0x00 0x00 0x00
__float______________________
0x7C 0xDC 0x8B 0xBE 0x0E 0x1C
__float______________________
0x88 0x33 0xFD 0x01 0xF1 0x45
__=6_____ __=3_____ __=5_____ __=12____
0x06 0x00 0x03 0x00 0x05 0x00 0x0C 0x00 0x00 0x00
                              __=1_____ __=1_____ __=10____ __=34____ __=0_____
0x00 0x00 0x00 0x00 0x00 0x00 0x01 0x00 0x01 0x00 0x0A 0x00 0x22 0x00 0x00 0x00
__=1_____
0x01 0x00 0x00 0x00 0x01 0x00 0x01 0x00 0x00 0x00 0x00 0x00 0x01 0x00 0x01 0x00
0x01 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x01 0x00 0x01 0x00 0x01 0x00 0x00 0x00
__=8_____ __=22____ __=255___ __=1_____ __=0_____ __=2_____ __=3_____ __=4_____
0x08 0x00 0x16 0x00 0xFF 0x00 0x01 0x00 0x00 0x00 0x02 0x00 0x03 0x00 0x04 0x00
__=9_____ __=6_____ __=10____ __=11____ __=12____ __=9_____ __=60____
0x09 0x00 0x06 0x00 0x0A 0x00 0x0B 0x00 0x0C 0x00 0x09 0x00 0x3C 0x00
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
0x83 0x00 0x00 0x00 0x00 0x70
__layer__ __size___
0x0B 0x00 0x11 0x00 0x0C 0x31 0x2D 0x4C 0x61 0x79 0x65 0x72

header end: 244, header length: 244
---house2.dch---------------------------------------------------------------
__=0________    ___=200_____    ___=2_______    ___=30______
0x00	0x00	0xc8	0x00	0x02	0x00	0x30	0x00
____________________________________________    ____________
0x81	0xef	0x3d	0x7c	0x2c	0xea	0x82	0x2f
____________________________
0x3f	0xcf	0x2e	0x99	0x88	0x00	0x00	0x00
0x00	0x48	0x86	0xe5	0x43	0x50	0x7d	0x12
0x81    0xdc    0x8b    0xbe    0x0e    0x1c    0x81    0xdc
0x8b    0xbe    0x0e    0x1c    0x7f    0xdc    0x8b    0xbe
0x0e    0x1c    0x7f    0xdc    0x8b    0xbe    0x0e    0x1c
___=3_______    ___=2_______    ___=1_______    ___=4_______
0x03    0x00    0x02    0x00    0x01    0x00    0x04    0x00
___=40______
0x28    0x00    0x7f    0xdc    0x8b    0xbe    0x0e    0x1c
0x7e    0xc7    0xdf    0x30    0xb1    0x79    0x81    0x9a
0x99    0x99    0x99    0x19    0x00    0x00    0x00    0x00
0x00    0x00    0x7e    0xdc    0x8b    0xbe    0x0e    0x1c
0x86    0xe5    0x43    0x50    0x7d    0x12    0x06    0x00
                ___=5_______    ___=12______
0x03    0x00    0x05    0x00    0x0c    0x00    0x00    0x00
0x00    0x00    0x00    0x00    0x00    0x00    0x01    0x00
                ___=10______    ___=34______
0x01    0x00    0x0a    0x00    0x22    0x00    0x00    0x00
0x01    0x00    0x00    0x00    0x01    0x00    0x01    0x00
0x00    0x00    0x00    0x00    0x01    0x00    0x01    0x00
0x01    0x00    0x00    0x00    0x00    0x00    0x01    0x00
0x01    0x00    0x00    0x00    0x00    0x00    0x01    0x00
___=8_______    ___=22______
0x08    0x00    0x16    0x00    0xff    0x00    0x01    0x00
0x00    0x00    0x0a    0x00    0x09    0x00    0x0b    0x00
0x0c    0x00    0x0d    0x00    0x0e    0x00    0x0f    0x00
___=8_______    ___=9_______
0x08    0x00    0x09    0x00    0x3c    0x00    0x83    0x00
0x00    0x00    0x00    0x70    0x83    0x00    0x00    0x00
0x00    0x70    0x83    0x00    0x00    0x00    0x00    0x70
0x83    0x00    0x00    0x00    0x00    0x70    0x83    0x00
0x00    0x00    0x00    0x70    0x83    0x00    0x00    0x00
0x00    0x70    0x83    0x00    0x00    0x00    0x00    0x70
0x83    0x00    0x00    0x00    0x00    0x70    0x83    0x00
0x00    0x00    0x00    0x70

Layer entry:

type: u16 = 11
length: u16 = 17
11 00 = 17
name: pascal string
01 00 0a 00
01 00 09 00
01 00 0b 00
01 00 0c 00
01 00 08 00

View entry:

type: u16 = 12
length: u16 = 33
name: pascal string
7d 2c f4 2f 68 ba 7e 4d 68 bb 6a 83 86 ad 48 93 c9 2f 85 f3 e5 05 58 79 0c 00 21 00


*/

static int
_parse(const char *data, size_t len, abydos_plugin_handle_t *h, cairo_t *cr)
{
    const char *end = data + len;
    /* skip pascal string, if present */
    if(data+1 > end){
        len = get8(data);
        if(len)
            data += 1 + len;
    }
    if(data + 86 > end)
        return -1;

    /* --- read header --- */
    /* check version */
    if(get16(data+2) > 200){
        len = 86;
        h->width = _get_double(data + 38);
        h->height = _get_double(data + 56);

        DEBUG_p("0: %f", _get_double(data + 6));
        DEBUG_p("1: %f (aspect ratio?)", _get_double(data + 16));
        DEBUG_p("w: %f", _get_double(data + 62));
        DEBUG_p("h: %f", _get_double(data + 80));
    }else{
        h->width = _get_double(data + 20);
        h->height = _get_double(data + 26);
        DEBUG_p("0: %f", _get_double(data + 8));
        DEBUG_p("1: %f", _get_double(data + 14));
        DEBUG_p("2: %f", _get_double(data + 32));
        DEBUG_p("3: %f", _get_double(data + 38));
        DEBUG_p("4: %f", _get_double(data + 44));
        len = 244;
    }
    DEBUG_p("size: %f x %f", h->width, h->height);

    data += len;
    if(data > end)
        return -1;

    for(;;){
        int type;
        if(data+4 > end)
            return 0;
        type = get16(data);
        len = get16(data+2);
        data += 4;
        //DEBUG_p("type: %d, length: %d", type, (int)len);
        if(data + len > end)
            return -1;
        switch(type){
        case 1: /* End */
            DEBUG_p("end");
            return 0;
        case 11: /* Layer */
            DEBUG_p("layer");
            break;
        case 12: /* View */
            DEBUG_p("view");
            break;
        case 17: /* Line */
            DEBUG_p("line");
            if(data + 28 > end)
                return -1;
            {
                //double d = *(double*)(data);
                DEBUG_p("%f", _get_double(data+8));
                DEBUG_p("move: %f x %f", _get_double(data+14), _get_double(data+20));
                DEBUG_p("line: %f x %f", _get_double(data+26), _get_double(data+32));
                cairo_move_to(cr, _get_double(data+14), _get_double(data+20));
                cairo_rel_line_to(cr, _get_double(data+26), _get_double(data+32));
                cairo_stroke(cr);
            }
            break;
        case 20: /* ??? */
            break;
        case 23: /* ??? */
            break;
        case 30: /* Box */
            break;
        default:
            DEBUG_p("unknown type %d of size %d", type, (int)len);
            break;
        }

        data += len;
    }
}

static int
_dcd_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    cairo_t *cr;
    int res;
    h->surface = cairo_recording_surface_create(CAIRO_CONTENT_COLOR, NULL);
    cr = cairo_create(h->surface);

    cairo_set_line_width(cr, 40);
    cairo_set_source_rgb(cr, 0, 1, 0);
    res = _parse(data, len, h, cr);
    if(!res){
        h->scale = 4000.0 / h->width;
        h->info->width = ceil(h->width * h->scale);
        h->info->height = ceil(h->height * h->scale);
    }

    cairo_destroy(cr);
    return res;
}


static void
_dcd_render(abydos_plugin_handle_t *h,
            cairo_t *cr,
            int page,
            cairo_bool_t *layer_visibility,
            int variant,
            double time,
            int frame)
{
    DEBUG_p("%f x %f", h->width * h->scale, h->height * h->scale);
    cairo_save(cr);
    cairo_scale(cr, h->scale, -h->scale);
    cairo_translate(cr, 0, -h->height);
    cairo_set_source_surface(cr, h->surface, 0, 0);
    cairo_paint(cr);
    cairo_restore(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _dcd_get_mime_types,
        .new = _dcd_new,
        .free = _dcd_free,
        .create_from_data = _dcd_create_from_data,
        .render = _dcd_render
    };
    return &p;
}
