/*
 * Copyright (C) 2018 - 2021  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <libnsbmp.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


#define ENABLE_CUR2ICO_HACK
#define ENABLE_WINICON /* Doesn't support the cursor subtype which unfortunately uses the same MIME type */

typedef struct{
    cairo_surface_t *surface;
    int width;
    int height;
}variant_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_BMP,
        T_ICO
    }type;
    union{
        cairo_surface_t *surface;
        variant_t *variant;
    };
};


static const char **
_nsbmp_get_mime_types()
{
    static const char *mime_types[] = {
        "image/bmp",
#ifdef ENABLE_WINICON
        "image/vnd.microsoft.icon",
#endif
        NULL
    };
    return mime_types;
}


static void *
_nsbmp_bitmap_create(int width, int height, unsigned int state)
{
    DEBUG_p("width: %d, height: %d, state: %d",width,height,state);
    return calloc(width * height, 4);
}


static unsigned char *
_nsbmp_bitmap_get_buffer(void *bitmap)
{
    return bitmap;
}


static size_t
_nsbmp_bitmap_get_bpp(void *bitmap)
{
    DEBUG_p("begin");
    return 4;
}


static abydos_plugin_handle_t *
_nsbmp_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_WINICON
    /* This block isn't needed if there can be only only one MIME type. */
    if(!strcasecmp(mime_type,"image/bmp")){
        h->type = T_BMP;
        h->surface = NULL;
    }else if(!strcasecmp(mime_type,"image/vnd.microsoft.icon")){
        h->type = T_ICO;
        h->variant = NULL;
    }else{
        /* not reached */
        free(h);
        return NULL;
    }
#else
    h->type = T_BMP;
#endif
    h->info = info;
    h->info->threadsafe = 1;
    return h;
}


static void
_nsbmp_free(abydos_plugin_handle_t *h)
{
    switch(h->type){
    case T_BMP:
        if(h->surface)
            cairo_surface_destroy(h->surface);
        break;
    case T_ICO:
        if(h->variant){
            int i;
            for(i=0;i<h->info->variant_count;++i)
                cairo_surface_destroy(h->variant[i].surface);
            free(h->variant);
        }
        break;
    }
    free(h);
}


static cairo_surface_t *
_surface_from_bmp(bmp_image* bmp){
    if(!bmp->decoded)
        if(bmp_decode(bmp)!=BMP_OK)
            return NULL;
    return nil_cairo_surface_create_from_direct(bmp->bitmap,bmp->width,bmp->height,-1,4,NILE_FORMAT_r8g8b8a8,0);
}


static int
_nsbmp_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    static bmp_bitmap_callback_vt vt = {
        .bitmap_create = _nsbmp_bitmap_create,
        .bitmap_destroy = free,
        .bitmap_get_buffer = _nsbmp_bitmap_get_buffer,
        .bitmap_get_bpp = _nsbmp_bitmap_get_bpp
    };
    bmp_image bmp;
    ico_collection ico;
#ifdef ENABLE_CUR2ICO_HACK
    char *buffer;
#endif
    const char *p;
    int ret = -1;
    switch(h->type){
    case T_BMP:
        bmp_create(&bmp,&vt);
        if(bmp_analyse(&bmp,len,(uint8_t*)data)!=BMP_OK){
            h->info->error = "Invalid BMP image";
            goto bmp_out;
        }
        h->surface = _surface_from_bmp(&bmp);
        if(!h->surface){
            h->info->error = "Decoding error";
            goto bmp_out;
        }
        h->info->width = bmp.width;
        h->info->height = bmp.height;
        ret = 0;
bmp_out:
        bmp_finalise(&bmp);
        break;
    case T_ICO:
#ifdef ENABLE_CUR2ICO_HACK
        if(len < 6)
            return -1;
        if(2==data[2]){
            p = buffer = (char*)malloc(len);
            memcpy(buffer,data,len);
            buffer[2] = 1;
        }else{
            buffer = NULL;
            p = data;
        }
#else
        p = data;
#endif
        ico_collection_create(&ico,&vt);
        if(ico_analyse(&ico,len,(uint8_t*)p)!=BMP_OK){
            h->info->error = "Invalid ICO image";
            goto ico_out;
        }
        {
            struct ico_image *iter;
            int i;
            h->info->variant_count = 0;
            for(iter=ico.first;iter;iter=iter->next)
                ++h->info->variant_count;
            h->variant = (variant_t*)malloc(h->info->variant_count * sizeof(variant_t));
            for(i=0,iter=ico.first;i<h->info->variant_count;++i,iter=iter->next){
                h->variant[i].surface = _surface_from_bmp(&iter->bmp);
                h->variant[i].width = iter->bmp.width;
                h->variant[i].height = iter->bmp.height;
                if(iter->bmp.width > h->info->width)
                    h->info->width = iter->bmp.width;
                if(iter->bmp.height > h->info->height)
                    h->info->height = iter->bmp.height;
            }
        }
        ret = 0;
ico_out:
        if(buffer)
            free(buffer);
        ico_finalise(&ico);
        break;
    }
    return ret;
#if 0
    h->info->frame_count = bmp.frame_count;
    h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        bmp_decode_frame(&bmp,i);
        /* FIXME!! Should we care if a frame fails to draw on a buffer
         * we know is valid? Will the next call to bmp_decode_frame()
         * be dangerous?
         */
         h->frame[i].surface = _surface_from_pixels(h, bmp.frame_image);
         h->frame[i].duration = (double)bmp.frames[i].frame_delay/100;
    }
#endif
}


static cairo_surface_t *
_nsbmp_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    switch(h->type){
    case T_BMP:
        return cairo_surface_reference(h->surface);
    case T_ICO:
        return cairo_surface_reference(h->variant[variant].surface);
    }
    return NULL;
}


static void
_nsbmp_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size){
    int i;
    for(i=0;i<h->info->variant_count;++i){
        size[i].width = h->variant[i].width;
        size[i].height = h->variant[i].height;
    }
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _nsbmp_get_mime_types,
        .new = _nsbmp_new,
        .free = _nsbmp_free,
        .create_from_data = _nsbmp_create_from_data,
        .get_image_surface = _nsbmp_get_image_surface,
        .get_variant_size = _nsbmp_get_variant_size
    };
    return &p;
}
