/*
 * Copyright (C) 2018 - 2021  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <libnsgif.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_nsgif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/gif",
        NULL
    };
    return mime_types;
}


static void *
_nsgif_bitmap_create(int width, int height)
{
    return calloc(width * height, 4);
}


static unsigned char *
_nsgif_bitmap_get_buffer(void *bitmap)
{
    return bitmap;
}


static abydos_plugin_handle_t *
_nsgif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_nsgif_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static int
_nsgif_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    static gif_bitmap_callback_vt vt = {
        .bitmap_create = _nsgif_bitmap_create,
        .bitmap_destroy = free,
        .bitmap_get_buffer = _nsgif_bitmap_get_buffer,
    };
    gif_animation gif;
    int code;
    int i;
    gif_create(&gif,&vt);
    do{
        code = gif_initialise(&gif,len,(unsigned char*)data);
    }while(GIF_WORKING == code);
    if(code != GIF_OK){
        /* FIXME!! set a more descriptive string */
        h->info->error = "Decoding error";
        goto out;
    }
    h->info->width = gif.width;
    h->info->height = gif.height;
    h->info->frame_count = gif.frame_count;
    h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        gif_decode_frame(&gif,i);
        /* FIXME!! Should we care if a frame fails to draw on a buffer
         * we know is valid? Will the next call to gif_decode_frame()
         * be dangerous?
         */
        h->frame[i].surface = nil_cairo_surface_create_from_direct(gif.frame_image,gif.width,gif.height,-1,4,NILE_FORMAT_r8g8b8a8,0);
        h->frame[i].duration = (double)gif.frames[i].frame_delay/100;
    }
out:
    gif_finalise(&gif);
    return GIF_OK == code ? 0 : -1;
}


static cairo_surface_t *
_nsgif_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_nsgif_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _nsgif_get_mime_types,
        .new = _nsgif_new,
        .free = _nsgif_free,
        .create_from_data = _nsgif_create_from_data,
        .get_image_surface = _nsgif_get_image_surface,
        .get_duration = _nsgif_get_duration
    };
    return &p;
}
