/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
extern "C"{
#define new new_
#include <abydos-plugin.h>
#undef new
#include <debug.h>
}
#include <QBuffer>
#include <QImageReader>


struct frame_t{
    cairo_surface_t *surface;
    double duration;
};

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static cairo_surface_t *
_surface_from_image(QImage *img)
{
    int width = img->width();
    int height = img->height();
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    for(int y=0;y<height;++y){
        for(int x=0;x<width;++x){
            QRgb p = img->pixel(x,y);
            int a = qAlpha(p);
            int r = qRed(p);
            int g = qGreen(p);
            int b = qBlue(p);
            *dst++ = (a*r/255)<<16|
                     (a*g/255)<< 8|
                     (a*b/255)<< 0|
                     a<<24;
        }
        dst += rowskip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_init_from_reader(abydos_plugin_handle_t *h,QImageReader *r)
{
    h->info->width = r->size().width();
    h->info->height = r->size().height();
    DEBUG_p("size: %dx%d",h->info->width,h->info->height);
    DEBUG_p("image count: %d",r->imageCount());
    h->info->frame_count = std::max(1,r->imageCount());
    h->frame = new frame_t[h->info->frame_count];
    for(int i=0;i<h->info->frame_count;++i){
        QImage img = r->read();
        if(img.width() > h->info->width)
            h->info->width = img.width();
        if(img.height() > h->info->height)
            h->info->height = img.height();
        h->frame[i].surface = _surface_from_image(&img);
        h->frame[i].duration = (double)r->nextImageDelay()/1000;
        r->jumpToNextImage();
    }
    return 0;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "application/x-krita",
                "image/openraster",
                //"image/x-eps",
                "image/x-rgb", "image/x-sgi",
                NULL
            };
            auto types1 = QImageReader::supportedImageFormats();
            for (auto const& i : types1) {
                DEBUG_p("%.*s",(int)i.length(),i.constData());
            }
            auto types2 = QImageReader::supportedMimeTypes();
            for (auto const& i : types2) {
                DEBUG_p("%.*s",(int)i.length(),i.constData());
            }

            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto *h = new abydos_plugin_handle_t;
            h->info = info;
            h->info->threadsafe = 1;
            h->frame = NULL;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->frame){
                for(int i=0;i<h->info->frame_count;++i)
                    cairo_surface_destroy(h->frame[i].surface);
                delete h->frame;
            }
            delete h;
        },
        .create_from_data = [](abydos_plugin_handle_t *h,const char *data,size_t len){
            QByteArray bytes(data,len);
            QBuffer buffer(&bytes);
            QImageReader r(&buffer);
            return _init_from_reader(h,&r);
        },
        .create_from_file = [](abydos_plugin_handle_t *h,const char *filename){
            QImageReader r(filename);
            return _init_from_reader(h,&r);
        },
        .get_image_surface = [](abydos_plugin_handle_t *h,
                                int page,
                                cairo_bool_t *layer_visibility,
                                int variant,
                                double time,
                                int frame)
        {
            return cairo_surface_reference(h->frame[frame].surface);
        },
        .get_duration = [](abydos_plugin_handle_t *h, int page, double *duration){
            for(int i=0;i<h->info->frame_count;++i)
                duration[i] = h->frame[i].duration;
        },
    };
    return &p;
}
