/*
 * Copyright (C) 2021 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <librosprite.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>


typedef struct{
    cairo_surface_t *surface;
    int width;
    int height;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_ros_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-acorn-sprite",
        "image/x-riscos-sprite",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ros_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    return h;
}


static void
_ros_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


static void
_page_from_sprite(page_t *page, struct rosprite *sprite)
{
    nile_t ni;

    if(sprite->mode.color_model != ROSPRITE_RGB)
        return;

    nile_init(&ni);

    page->width = ni.width = sprite->width;
    page->height = ni.height = sprite->height;
    ni.type = NILE_TYPE_DIRECT;
    ni.pixels = (uint8_t*)sprite->image;
#if __BYTE_ORDER == __BIG_ENDIAN
    ni.format = NILE_FORMAT_b8g8r8a8;
#else
    ni.format = NILE_FORMAT_a8r8g8b8;
#endif
    ni.bpc = 4;
    page->surface = nil_cairo_surface_create(&ni, 0);

    nile_done(&ni);
}


static void
_handle_area(abydos_plugin_handle_t *h, struct rosprite_area *area)
{
    int i;
    h->info->page_count = area->sprite_count;
    DEBUG_p("%d pages", h->info->page_count);
    h->page = calloc(h->info->page_count, sizeof(page_t));
    for(i=0;i<h->info->page_count;++i)
        _page_from_sprite(&h->page[i], area->sprites[i]);
}


static int
_ros_create_from_handle(abydos_plugin_handle_t *h, FILE *f)
{
    struct rosprite_file_context *ctx;
    struct rosprite_area *area;
    //rosprite_error err;

    if(rosprite_create_file_context(f, &ctx) != ROSPRITE_OK){
        return -1;
    }

    if(rosprite_load(rosprite_file_reader, ctx, &area) != ROSPRITE_OK){
        DEBUG_p("error loading spritefile\n");
        return -1;
    }

    _handle_area(h, area);

    rosprite_destroy_file_context(ctx);
    return 0;
}


static int
_ros_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    struct rosprite_mem_context *ctx;
    struct rosprite_area *area;
    //rosprite_error err;

    if(rosprite_create_mem_context((uint8_t*)data, len, &ctx) != ROSPRITE_OK){
        return -1;
    }

    if(rosprite_load(rosprite_file_reader, ctx, &area) != ROSPRITE_OK){
        DEBUG_p("error loading spritefile\n");
        return -1;
    }

    _handle_area(h, area);

    rosprite_destroy_mem_context(ctx);
    return 0;
}


#if 1
static int
_ros_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    FILE *f = fopen(filename, "r");
    int res;
    if(!f)
        return -1;
    res = _ros_create_from_handle(h, f);
    fclose(f);
    return res;
}
#endif


static cairo_surface_t *
_ros_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


static void
_ros_get_page_info(abydos_plugin_handle_t *h,
                   int page,
                   abydos_plugin_page_info_t *info)
{
    info->width = h->page[page].width;
    info->height = h->page[page].height;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ros_get_mime_types,
        .new = _ros_new,
        .free = _ros_free,
        .create_from_data = _ros_create_from_data,
        .create_from_file = _ros_create_from_file,
        .get_image_surface = _ros_get_image_surface,
        .get_page_info = _ros_get_page_info
    };
    return &p;
}
