/*
 * Copyright (C) 1995 - 1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 2009         Mica, Östersunds kummun
 * Copyright (C) 2016 - 2020  Magnus Bergman
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/*
 * Based on ghash.c from GLIB (ftp://ftp.gtk.org/pub/gtk/)
 */

#include <ctype.h>                      /* tolower() */
#include <limits.h>                     /* UINT_MAX */
#include <stdint.h>
#include <stdlib.h>                     /* calloc(), free() */
#include <string.h>                     /* memset(), strcmp() */

#include "hashmap.h"


#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define HASH_MAP_MIN_SHIFT 3            /* 1 << 3 == 8 buckets */
#define UNUSED_HASH 0
#define TOMBSTONE_HASH 1
#define FIRST_REAL_HASH 2
#define HASH_IS_REAL(a) ((a)>=FIRST_REAL_HASH)

/* --- Hash Functions --- */

hash_t hash_combine(hash_t a, hash_t b){
    a ^= b + 0x9e3779b9 + (a << 6) + (a >> 2);
    return a;
}


int direct_equal(const void* a,const void* b){
    return a==b;
}

hash_t direct_hash(const void* a){
    return (intptr_t)a;
}


hash_t pointer_hash(const void* a){
#if 1
    hash_t h = (intptr_t)a;
#else
# if INTPTR_MAX == INT64_MAX
    /* simply strip the MSBs if sizeof(void*) > sizeof(int) */
    unsigned int h = ((uint64_t)a) & UINT_MAX;
# else
    unsigned int h = (uint32_t)a;
# endif
#endif
#if 1
    h = ~h + (h << 15);
    h = h ^ (h >> 12);
    h = h + (h << 2);
    h = h ^ (h >> 4);
    h = h * 2057;
    h = h ^ (h >> 16);
#endif
    return h;
}


int int_equal(const void* a,const void* b){
    return *((int*)a) == *((int*)b);
}

hash_t int_hash(const void* a){
    hash_t h = *((int*)a);
    h = ((h >> 16) ^ h) * 0x45d9f3b;
    h = ((h >> 16) ^ h) * 0x45d9f3b;
    h = ((h >> 16) ^ h);
    return h;
}


int str_equal(const void* a,const void* b){
    return !strcmp((char*)a,(char*)b);
}

hash_t str_hash_update(hash_t h,const char* s,int len){
    const signed char* p = (const signed char*)s;
    while(*p && len){
        h = (h<<5)-h+*p;
        ++p;
        --len;
    }
    return h;
}

hash_t str_hash(const void* a){
    /* 31 bit hash function */
    const signed char* p = a;
    hash_t h = *p;
    return str_hash_update(h,((const char*)a)+1,-1);
}


int strcase_equal(const void* a,const void* b){
    return !strcasecmp((char*)a,(char*)b);
}

hash_t strcase_hash_update(hash_t h,const char* s,int len){
    const signed char* p = (const signed char*)s;
    while(*p && len){
        h = (h<<5)-h+tolower(*p);
        ++p;
        --len;
    }
    return h;
}

hash_t strcase_hash(const void* a){
    /* 31 bit hash function */
    const signed char* p = a;
    hash_t h = *p;
    return strcase_hash_update(h,((const char*)a)+1,-1);
}

/* --- private functions ------------------------------------------------ */

static const int prime_mod [] = {
    1,          /* For 1 << 0 */
    2,
    3,
    7,
    13,
    31,
    61,
    127,
    251,
    509,
    1021,
    2039,
    4093,
    8191,
    16381,
    32749,
    65521,      /* For 1 << 16 */
    131071,
    262139,
    524287,
    1048573,
    2097143,
    4194301,
    8388593,
    16777213,
    33554393,
    67108859,
    134217689,
    268435399,
    536870909,
    1073741789,
    2147483647  /* For 1 << 31 */
};


static void
_hashmap_set_shift(hashmap_t *hm, int shift)
{
    unsigned int mask = 0;
    int i;
    hm->size = 1 << shift;
    hm->nodes = calloc(hm->size, sizeof(hashmap_node_t));
    hm->mod  = prime_mod[shift];
    for(i=0;i<shift;++i){
        mask <<= 1;
        mask |= 1;
    }
    hm->mask = mask;
}


static int
_hashmap_find_closest_shift(int n)
{
    int i;
    for(i=0;n;++i)
        n >>= 1;
    return i;
}


static hashmap_node_t *
_hashmap_lookup_node(hashmap_t *hm, const void* key, hash_t *r_hash)
{
    unsigned int index;
    hashmap_node_t *first_tombstone = NULL; /* Just to silent compiler warnings. */
    int have_tombstone = 0;
    unsigned int step = 0;
    hash_t hash = hm->hash_func(key);
    if(!HASH_IS_REAL(hash))
        hash = FIRST_REAL_HASH;

    if(r_hash)
        *r_hash = hash;

    for(index=hash%hm->mod;;){
        hashmap_node_t *node = &hm->nodes[index];
        if(UNUSED_HASH == node->hash){
            return have_tombstone?first_tombstone:node;
        }else if(node->hash == hash){
            if(hm->equal_func){
                if(hm->equal_func(node->key,key))
                    return node;
            }else if(node->key == key){
                return node;
            }
        }else if(TOMBSTONE_HASH == node->hash && !have_tombstone){
            first_tombstone = node;
            have_tombstone = 1;
        }
        index += ++step;
        index &= hm->mask;
    }
}


static void
_hashmap_remove_all_nodes(hashmap_t *hm, int destruction){
    if(hm->nnodes > 0 && (hm->key_notify_func || hm->value_notify_func)){
        int i;
        for(i=0;i<hm->size;++i){
            hashmap_node_t *node = &hm->nodes[i];
            if(HASH_IS_REAL(node->hash)){
                if(hm->key_notify_func)
                    hm->key_notify_func(node->key);
                if(hm->value_notify_func)
                    hm->value_notify_func(node->value);
            }
        }
    }

    if(destruction){
        free(hm->nodes);
        return;
    }

    if(hm->size <= (1<<HASH_MAP_MIN_SHIFT)){
        /* No need to downsize, just clear it */
        memset(hm->nodes, 0, hm->size * sizeof(hashmap_node_t));
    }else{
        free(hm->nodes);
        _hashmap_set_shift(hm, HASH_MAP_MIN_SHIFT);
    }
    hm->nnodes = hm->noccupied = 0;
}


static void
_hashmap_resize(hashmap_t *hm)
{
    hashmap_node_t *old_nodes;
    int old_size;

    old_nodes = hm->nodes;
    old_size = hm->size;
    _hashmap_set_shift(hm,MAX(_hashmap_find_closest_shift(hm->nnodes*2),HASH_MAP_MIN_SHIFT));

    if(hm->nnodes > 0){
        int i;
        for(i=0;i<old_size;++i){
            hashmap_node_t *old_node = &old_nodes[i];
            hashmap_node_t *new_node;
            unsigned int index;
            unsigned int step = 0;

            if(!HASH_IS_REAL(old_node->hash))
                continue;

            for(index=old_node->hash%hm->mod;;){
                new_node = &hm->nodes[index];
                if(!new_node->hash)
                    break;
                index += ++step;
                index &= hm->mask;
            }
            *new_node = *old_node;
        }
    }
    free(old_nodes);
    hm->noccupied = hm->nnodes;
}


static inline void
_hashmap_maybe_resize(hashmap_t *hm){
    int noccupied = hm->noccupied;
    int size = hm->size;
    if((size > hm->nnodes*4 && size > 1<<HASH_MAP_MIN_SHIFT) || (size <= noccupied+(noccupied/16)))
        _hashmap_resize(hm);
}

/* --- public functions ------------------------------------------------- */

hashmap_t *
hashmap_init(hashmap_t *hm,
             hashmap_hash_func_t hash_func,
             hashmap_equal_func_t equal_func,
             hashmap_notify_func_t key_notify_func,
             hashmap_notify_func_t value_notify_func)
{
    _hashmap_set_shift(hm,HASH_MAP_MIN_SHIFT);
    hm->nnodes = 0;
    hm->noccupied = 0;
    hm->hash_func = hash_func ? hash_func : direct_hash;
    hm->equal_func = equal_func ? equal_func : direct_equal;
    hm->key_notify_func = key_notify_func;
    hm->value_notify_func = value_notify_func;
    return hm;
}


void
hashmap_done(hashmap_t *hm)
{
    _hashmap_remove_all_nodes(hm,1);
}


void
hashmap_add(hashmap_t *hm, void *key, void *value)
{
    unsigned int hash;
    hashmap_node_t *node = _hashmap_lookup_node(hm,key,&hash);
    hash_t old_hash = node->hash;

    if(HASH_IS_REAL(old_hash)){
        if(hm->key_notify_func)
            hm->key_notify_func(key);
        if(hm->value_notify_func)
            hm->value_notify_func(node->value);
        node->value = value;
    }else{
        node->key = key;
        node->value = value;
        node->hash = hash;

        ++hm->nnodes;

        if(!old_hash){
            /* We replaced an empty node, and not a tombstone */
            ++hm->noccupied;
            _hashmap_maybe_resize(hm);
        }
    }
}


static const hashmap_node_t *
_hash_map_get_node(hashmap_t *hm, const void *key)
{
    unsigned int index;
    unsigned int hash;
    int step = 0;

    hash = hm->hash_func(key);
    if(!HASH_IS_REAL(hash))
        hash = FIRST_REAL_HASH;

    for(index=hash%hm->mod;;){
        hashmap_node_t *node = &hm->nodes[index];
        if(UNUSED_HASH == node->hash){
            return NULL;
        }else if(node->hash == hash){
            if(hm->equal_func){
                if(hm->equal_func(node->key,key))
                    return node;
            }else if(node->key == key){
                return node;
            }
        }
        index += ++step;
        index &= hm->mask;
    }
}


void *
hashmap_get(hashmap_t *hm, const void *key)
{
    const hashmap_node_t *node = _hash_map_get_node(hm, key);
    return node ? node->value : NULL;
}


const void *
hashmap_get_key(hashmap_t *hm, const void *key)
{
    const hashmap_node_t *node = _hash_map_get_node(hm, key);
    return node ? node->key : NULL;
}


void *
hashmap_get_any(hashmap_t *hm)
{
    int i;
    for(i=0;i<hm->size;++i){
        hashmap_node_t *node = &hm->nodes[i];
        if(HASH_IS_REAL(node->hash))
            return node->value;
    }
    return NULL;
}


int
hashmap_remove(hashmap_t *hm, const void *key)
{
    hash_t hash;
    hashmap_node_t *node = _hashmap_lookup_node(hm,key,&hash);
    if(!HASH_IS_REAL(node->hash))
        return 0;

    /* Erect tombstone */
    node->hash = TOMBSTONE_HASH;

    --hm->nnodes;

    if(hm->key_notify_func)
        hm->key_notify_func(node->key);
    if(hm->value_notify_func)
        hm->value_notify_func(node->value);

    _hashmap_maybe_resize(hm);

    return 1;
}


void
hashmap_remove_all(hashmap_t *hm)
{
    _hashmap_remove_all_nodes(hm,0);
}



int
hashmap_foreach(hashmap_t *hm,hashmap_foreach_func_t func,void *userdata)
{
    int i;
    for(i=0;i<hm->size;++i){
        hashmap_node_t *node = &hm->nodes[i];
        if(HASH_IS_REAL(node->hash)){
            int ret = func(node->key,node->value,userdata);
            if(ret)
                return ret;
        }
    }
    return 0;
}
