/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <math.h>

#include "path.h"


static void
_weight_points2(loa_point_t *p, const loa_point_t *p1, const loa_point_t *p2, double frac)
{
    p->x = p1->x + (p2->x - p1->x) * frac;
    p->y = p1->y + (p2->y - p1->y) * frac;
}


double
_curve_length(const loa_point_t *start, const loa_point_t *c2, const loa_point_t *c5, const loa_point_t *end, int splits)
{
    if(splits){
        loa_point_t c1;
        loa_point_t c3;
        loa_point_t c4;
        loa_point_t c6;
        loa_point_t mid;
        _weight_points2(&c1, start, c2, 0.5);
        _weight_points2(&c6, c5, end, 0.5);
        _weight_points2(&mid, c2, c5, 0.5);
        _weight_points2(&c3, &c1, &mid, 0.5);
        _weight_points2(&c4, &mid, &c6, 0.5);
        _weight_points2(&mid, &c3, &c4, 0.5);
        return _curve_length(start, &c1, &c3, &mid, splits-1) + _curve_length(&mid, &c4, &c6, end, splits-1);
    }else{
        return sqrt(pow(end->x - start->x, 2) + pow(end->y - start->y, 2));
    }
}


double
path_segment_length(const loa_segment_t *seg, const loa_point_t *point, path_state_t *ps)
{
    const loa_point_t *p;
    switch(seg->type){
    case LOA_GOTO:
        ps->close = ps->next = point[seg->point];
        return 0;
    case LOA_LINE:
        p = &point[seg->point];
        ps->next = p[0];
        return sqrt(pow(p[0].x - ps->last.x, 2) + pow(p[0].y - ps->last.y, 2));
    case LOA_CURVE:
        p = &point[seg->point];
        ps->next = p[2];
        return _curve_length(&ps->last, &p[0], &p[1], &p[2], 4);
    case LOA_CLOSE:
        return sqrt(pow(ps->close.x - ps->last.x, 2) + pow(ps->close.y - ps->last.y, 2));
    }
    /* not reached */
    return 0;
}


double
path_length(const loa_path_t *path, const loa_segment_t *segment, const loa_point_t *point)
{
    path_state_t ps;
    double length = 0;
    int i;
    for(i=0;i<path->segment_count;++i){
        const loa_segment_t *seg = &segment[path->segment+i];
        length += path_segment_length(seg, point, &ps);
        DEBUG_p("after segment %d: %f", i, length);
        ps.last = ps.next;
    }
    return length;
}


void
path_curve_trace(const loa_point_t *p1, const loa_point_t *p2, const loa_point_t *p3, const loa_point_t *p4, double frac, double *x, double *y, double *a)
{
    loa_point_t c1, c3, c4, c6, mid;
    double dx, dy;
    _weight_points2(&c1, p1, p2, frac);
    _weight_points2(&c6, p3, p4, frac);
    _weight_points2(&mid, p2, p3, frac);
    _weight_points2(&c3, &c1, &mid, frac);
    _weight_points2(&c4, &mid, &c6, frac);
    dx = c4.x - c3.x;
    dy = c4.y - c3.y;
    *x = c3.x + dx * frac;
    *y = c3.y + dy * frac;
    if(a)
        *a = atan2(dy, dx);
}


void
path_segment_trace(const loa_segment_t *seg, const loa_point_t *point, path_state_t *ps, double frac, double *x, double *y, double *a)
{
    const loa_point_t *p;
    double dx, dy;
    switch(seg->type){
    case LOA_GOTO:
        break;
    case LOA_LINE:
        p = &point[seg->point];
        dx = p[0].x - ps->last.x;
        dy = p[0].y - ps->last.y;
        *x = ps->last.x + dx * frac;
        *y = ps->last.y + dy * frac;
        break;
    case LOA_CURVE:
        p = &point[seg->point];
#if 1
        path_curve_trace(&ps->last, &p[0], &p[1], &p[2], frac, x, y, a);
        return;
#else
        {
            loa_point_t c1;
            loa_point_t c3;
            loa_point_t c4;
            loa_point_t c6;
            loa_point_t mid;
            _weight_points2(&c1, &ps->last, &p[0], frac);
            _weight_points2(&c6, &p[1], &p[2], frac);
            _weight_points2(&mid, &p[0], &p[1], frac);
            _weight_points2(&c3, &c1, &mid, frac);
            _weight_points2(&c4, &mid, &c6, frac);
            dx = c4.x - c3.x;
            dy = c4.y - c3.y;
            *x = c3.x + dx * frac;
            *y = c3.y + dy * frac;
            *a = atan2(dy, dx);
        }
        break;
#endif
    case LOA_CLOSE:
        dx = ps->close.x - ps->last.x;
        dy = ps->close.y - ps->last.y;
        *x = ps->last.x + dx * frac;
        *y = ps->last.y + dy * frac;
        break;
    }
    if(a)
        *a = atan2(dy, dx);
}


void
path_trace(const loa_path_t *path, double length, const loa_segment_t *segment, const loa_point_t *point, double frac, double *x, double *y, double *a)
{
    path_state_t ps;
    double pos_target = length * frac;
    double pos_last = 0;
    int i;
    for(i=0;i<path->segment_count;++i){
        double pos_next;
        const loa_segment_t *seg = &segment[path->segment+i];
        pos_next = pos_last + path_segment_length(seg, point, &ps);
        if(pos_next > pos_target){
            path_segment_trace(seg, point, &ps, (pos_target-pos_last)/(pos_next-pos_last), x, y, a);
            return;
        }
        pos_last = pos_next;
        ps.last = ps.next;
    }
}


void
path_start(const loa_path_t *path, const loa_segment_t *segment, const loa_point_t *point, double *x, double *y)
{
    const loa_segment_t *seg = &segment[path->segment];
    const loa_point_t *p = &point[seg->point];
    *x = p->x;
    *y = p->y;
}
