/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <ctype.h>
#include <debug.h>
#include <loa-buffer.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "hashmap.h"
#include "path.h"
#include "svgcolor.h"
#include "xmlutil.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#define POINTER2INT(a) ((int)(intptr_t)(a))
#define INT2POINTER(a) ((void*)(intptr_t)(a))

/* The maximum angle of an arc to convert to one curve. */
#define MAX_ANGLE (M_PI/2)

typedef struct{
    color_t color;
    double frac;
}gradient_stop_t;

typedef struct{
    int first;
    int last;
}gradient_t;

typedef struct{
    enum{
        PT_NONE,
        PT_COLOR,
        PT_GRADIENT
    }type;
    union{
        color_t color;
        gradient_t gradient;
    };
}pattern_t;

typedef struct{
    color_t fill_color;
    color_t stroke_color;
    color_t current_color;
    double stroke_width;
}style_t;

struct _loa_object_t{
    enum{
        OT_GROUP,
        OT_ELLIPSE,
        OT_PATH
    }type;
    style_t style;
    /* points to define the shape */
    int first;
    int last;
    union{
        struct{
            double rw;
            double rh;
        }ellipse;
        loa_path_t path;
    };
};

struct _loa_keyframe_t{
    union{
        double value[3];
        struct{
            double dx;
            double dy;
        }move;
        struct{
            double a;
            double x;
            double y;
        }rotate;
        struct{
            double s;
            double x;
            double y;
        }scale;
        double alpha;
        cairo_matrix_t matrix;
    };
    union{
        struct{
            loa_point_t c1;
            loa_point_t c2;
        }spline;
    };
    double frac;
};

struct _loa_anim_t{
    enum{
        AT_MOVE,
        AT_ROTATE,
        AT_SCALE,
        AT_MOTION,
        AT_FILL_ALPHA,
        AT_STROKE_DASHARRAY,
        AT_STROKE_DASHOFFSET
    }type;
    enum{
        AM_DISCRETE,
        AM_LINEAR,
        AM_PACED,
        AM_SPLINE
    }mode;
    union{
        struct{
            int first;
            int last;
        }point;
        struct{
            int first;
            int last;
        }shape;
    };
    union{
        struct{
            int first;
            int last;
        }keyframe;
        struct{
            loa_path_t path;
            double length;
        }motion;
    };
    double duration;
};


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    loa_point_t *point;
    int point_count;
    loa_segment_t *segment;
    int segment_count;
    loa_object_t *shape;
    int shape_count;
    loa_anim_t *anim;
    int anim_count;
    loa_keyframe_t *keyframe;
    int keyframe_count;
};


static const char **
_svg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/svg+xml",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_svg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->scalable_size = 1;
    h->info->scalable_time = 1;
    h->info->threadsafe = 1;
    h->point = NULL;
    h->segment = NULL;
    h->shape = NULL;
    h->anim = NULL;
    return h;
}


static void
_svg_free(abydos_plugin_handle_t *h)
{
    if(h->point)
        free(h->point);
    if(h->segment)
        free(h->segment);
    if(h->shape)
        free(h->shape);
    if(h->anim)
        free(h->anim);
    free(h);
}


static void
_get_matrix(cairo_matrix_t *m, const char **s)
{
    m->xx = _get_double(s);
    m->yx = _get_double(s);
    m->xy = _get_double(s);
    m->yy = _get_double(s);
    m->x0 = _get_double(s);
    m->y0 = _get_double(s);
}


static void
_matrix_multiply_string(cairo_matrix_t *m, const char *s)
{
    if(!s)
        return;
    DEBUG_p("%s", s);
    if(!strncmp(s,"matrix(",7)){
        cairo_matrix_t m0;
        s += 7;
        _get_matrix(&m0, &s);
        cairo_matrix_multiply(m, &m0, m);
        return;
    }else if(!strncmp(s,"rotate(",7)){
        double a, x, y;
        s += 7;
        a = _get_double(&s) * M_PI / 180;
        x = _get_double(&s);
        y = _get_double(&s);
        //cairo_matrix_translate(m, x, -y);
        cairo_matrix_translate(m, x, y);
        //cairo_matrix_translate(m, 400, 300);
        cairo_matrix_rotate(m, a);
    }else if(!strncmp(s,"translate(",10)){
        double x;
        double y;
        s += 10;
        x = _get_double(&s);
        y = _get_double(&s);
        cairo_matrix_translate(m, x, y);
        return;
    }
}


static void
_color_set_string(svg_color_parser_t *p, color_t *c, const char *s, int len, color_t *current)
{
    c->a = 1;
#if 1
    if(!strncasecmp(s, "none", 4)){
        c->r = -1;
    }else if(!strncasecmp(s, "currentColor", 12)){
        DEBUG_p("current color");
        c->r = current->r;
        c->g = current->g;
        c->b = current->b;
    }else if(!strncasecmp(s, "inherit", 7)){
        DEBUG_p("inherit color");
    }else{
        svg_color_parse(p, c, s, len);
    }
#else
    if('#'==*s || isdigit(*s)){
        _color_set_hexstring(c, s);
    }else if(!strncasecmp(s, "currentColor", 12)){
        DEBUG_p("current color");
        c->r = current->r;
        c->g = current->g;
        c->b = current->b;
    }else if(!strncasecmp(s, "inherit", 7)){
        DEBUG_p("inherit color");
    }else if(!strncasecmp(s, "none", 4)){
        c->r = -1;
    }else{
        const svg_color_t *sc = svg_color_lookup(s);
        if(sc){
            c->r = (double)((sc->rgb>>16)&0xff)/255;
            c->g = (double)((sc->rgb>> 8)&0xff)/255;
            c->b = (double)((sc->rgb>> 0)&0xff)/255;
        }else{
            c->r = 1;
            c->g = 0;
            c->b = 1;
        }
    }
#endif
}


typedef void(*style_set_func_t)(svg_color_parser_t *p, style_t*, const char*, int);


static void
_style_set_color(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    if(len < 0)
        len = strlen(s);
    _color_set_string(p, &st->current_color, s, len, &st->current_color);
}


static void
_style_set_fill(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    _color_set_string(p, &st->fill_color, s, len, &st->current_color);
}


static void
_style_set_fill_opacity(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    st->fill_color.a = strtod(s, NULL);
}


static void
_style_set_stroke(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    _color_set_string(p, &st->stroke_color, s, len, &st->current_color);
}


static void
_style_set_stroke_dasharray(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
}


static void
_style_set_stroke_opacity(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    st->stroke_color.a = strtod(s, NULL);
}


static void
_style_set_stroke_width(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    st->stroke_width = strtod(s, NULL);
}


static void
_style_set_string(svg_color_parser_t *p, style_t *st, const char *s, int len)
{
    static const struct{
        const char *s;
        int len;
        style_set_func_t func;
    }item[] = {
        {"color:", 6, _style_set_color},
        {"fill:", 5, _style_set_fill},
        {"fill-opacity:", 13, _style_set_fill_opacity},
        {"stroke:", 7, _style_set_stroke},
        {"stroke-dasharray:", 17, _style_set_stroke_dasharray},
        {"stroke-opacity:", 15, _style_set_stroke_opacity},
        {"stroke-width:", 15, _style_set_stroke_width},
        {NULL}
    };
    static const color_t default_color = {1,1,1,1};
    const char *end;
    if(!s)
        return;
    if(len<0)
        len = strlen(s);
    end = s+len;
    st->fill_color = default_color;
    while(s<end){
#if 1
        int i;
        for(i=0;item[i].s;++i){
            if(s+item[i].len <= end &&!strncmp(s, item[i].s, item[i].len)){
                s += item[i].len;
                item[i].func(p, st, s, end-s);
            }
        }
#else
        if(s+5 <= end && !strncmp(s, "fill:", 5)){
            s += 5;
            if(s+7 > end)
                break;
            _color_set_string(&st->fill_color,s);
        }else if(s+13 <= end && !strncmp(s, "fill-opacity:", 13)){
            s += 13;
            st->fill_color.a = strtod(s, NULL);
        }else if(s+7 <= end && !strncmp(s, "stroke:", 7)){
            s += 7;
            if(s+7 > end)
                break;
            _color_set_string(&st->stroke_color,s);
        }else if(s+17 <= end && !strncmp(s, "stroke-dasharray:", 17)){
        }else if(s+15 <= end && !strncmp(s, "stroke-opacity:", 15)){
            s += 15;
            st->stroke_color.a = strtod(s, NULL);
        }else if(s+13 <= end && !strncmp(s, "stroke-width:", 13)){
            s += 13;
            st->stroke_width = strtod(s, NULL);
        }
#endif
        while(s<end && *s!=';')
            ++s;
        while(s<end && (' '==*s||';'==*s))
            ++s;
    }
}


static void
_style_set_attributes(svg_color_parser_t *p, style_t *st, xml_attributes_t *attr)
{
    static const struct{
        const char *s;
        style_set_func_t func;
    }item[] = {
        {"color", _style_set_color},
        {"fill", _style_set_fill},
        {"fill-opacity", _style_set_fill_opacity},
        {"stroke", _style_set_stroke},
        {"stroke-dasharray", _style_set_stroke_dasharray},
        {"stroke-opacity", _style_set_stroke_opacity},
        {"stroke-width", _style_set_stroke_width},
        {"style", _style_set_string},
        {NULL}
    };
    const char *s;
#if 1
    int i;
    for(i=0;item[i].s;++i){
        s = xml_parser_attr_string(attr, item[i].s);
        if(s)
            item[i].func(p, st, s, -1);
    }
#else
    s = xml_parser_attr_string(attr, "fill");
    if(s)
        _color_set_string(&st->fill_color, s);
    s = xml_parser_attr_string(attr, "stroke");
    if(s)
        _color_set_string(&st->stroke_color, s);
    st->stroke_width = xml_parser_attr_double(attr, "stroke-width", 0);
    _style_set_string(st, xml_parser_attr_string(attr, "style"), -1);
#endif
}


static void
_arc_center(loa_point_t *center, double x1, double y1, double x2, double y2, double rx, double ry, double rot, int large, int sweep)
{
    cairo_matrix_t m;
    cairo_matrix_init_rotate(&m, rot);
    double x1p = (x1 - x2) / 2;
    double y1p = (y1 - y2) / 2;
    cairo_matrix_transform_point(&m, &x1p, &y1p);

    DEBUG_p("start: %fx%f, end: %fx%f",x1,y1,x2,y2);
    DEBUG_p("radius: %fx%f",rx,ry);
    //rx = abs(rx);
    //ry = abs(ry);
    // Step (c): Ensure radii are large enough
    double lambda = (x1p * x1p) / (rx * rx) + (y1p * y1p) / (ry * ry);
    DEBUG_p("lambda: %f",lambda);
    if(lambda > 1){
        rx = sqrt(lambda) * rx;
        ry = sqrt(lambda) * ry;
    }

    // Step 2: Compute (cx′, cy′)
    int sign = (large == sweep)? -1 : 1;
    double div = (rx * rx * ry * ry - rx * rx * y1p * y1p - ry * ry * x1p * x1p) /
                (rx * rx * y1p * y1p + ry * ry * x1p * x1p);
    double co = sign * sqrt(abs(div));

    // Step 3: Compute (cx, cy) from (cx′, cy′)
    cairo_matrix_invert(&m);
    center->x = (rx*y1p/ry)*co;
    center->y = (-ry*x1p/rx)*co;
    cairo_matrix_transform_point(&m, &center->x, &center->y);
    center->x += (x1+x2)/2;
    center->y += (y1+y2)/2;
    DEBUG_p("center: %fx%f",center->x,center->y);
}


static void
_loa_buffer_push_arc(loa_buffer_t *vb, cairo_matrix_t *m, loa_path_t *path, loa_point_t *last, double rx, double ry, double rot, int large, int sweep, double x, double y)
{
#if 1
    loa_point_t center;
    double dx;
    double dy;
    double a;
    double da;
    int steps;
    double c;
    int i;
    _arc_center(&center, last->x, last->y, x, y, rx, ry, rot, large, sweep);

    dx = (last->x - center.x) / rx;
    dy = (last->y - center.y) / ry;
    a = atan2(dy,dx);
    DEBUG_p("start angle: %f",180*a/M_PI);
    dx = (x - center.x) / rx;
    dy = (y - center.y) / ry;
    da = atan2(dy,dx) - a;
    if(!large && fabs(da)>M_PI){
        if(da>0){
            da -= 2*M_PI;
        }else{
            da += 2*M_PI;
        }
    }
    //if(large==sweep)
    //    da *= -1;
    DEBUG_p("arc angle: %f, large: %s, sweep: %s",180*da/M_PI,large?"YES":"NO",sweep?"YES":"NO");
    steps = ceil(fabs(da) / MAX_ANGLE);
    da /= steps;
    c = 4.0/3*tan(da/4);
    //if(large==sweep)
    //    c *= -1;
    for(i=0;i<steps;++i){
        loa_point_t *p = loa_buffer_push_segment(vb, path, LOA_CURVE);
        p[0].x = last->x - rx*c*sin(a);
        p[0].y = last->y + ry*c*cos(a);
        a += da;
#if 0
        p[2].x = center.x + rx*cos(a);
        p[2].y = center.y + ry*sin(a);
#else
        p[2].x = x;
        p[2].y = y;
#endif
        p[1].x = p[2].x + rx*c*sin(a);
        p[1].y = p[2].y - ry*c*cos(a);
        *last = p[2];
        //transform_apply_pos(t, &p[0].x, &p[0].y);
        //transform_apply_pos(t, &p[1].x, &p[1].y);
        //transform_apply_pos(t, &p[2].x, &p[2].y);
        cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
        cairo_matrix_transform_point(m, &p[1].x, &p[1].y);
        cairo_matrix_transform_point(m, &p[2].x, &p[2].y);
    }
#else
    int sign = (large==sweep)?-1:1;
    double dx;
    double dy;
    double a1;
    double a2;
    loa_point_t center;
    _arc_center(&center, last->x, last->y, x, y, rx, ry, rot, large, sweep);

    dx = (last->x - center.x) / rx;
    dy = (last->y - center.y) / ry;
    a1 = atan2(dy,dx);
    dx = (x - center.x) / rx;
    dy = (y - center.y) / ry;
    a2 = atan2(dy,dx);

    if(sign>0){
        while(a2 < a1)
            a2 += 2*M_PI;
    }else{
        while(a2 > a1)
            a1 += 2*M_PI;
    }

    while(sign*a1 < sign*a2){
        point_t *p = loa_buffer_push_segment(vb, path, LOA_CURVE);
        double da;
        double c;
        da = MIN(a2-a1, M_PI/2);
        c = 4.0/3*tan(da/4);

        p[0].x = last->x - rx*c*sin(a1);
        p[0].y = last->y + ry*c*cos(a1);
        a1 += da;
#if 0
        p[2].x = center.x + rx*cos(a1);
        p[2].y = center.y + ry*sin(a1);
#else
        p[2].x = x;
        p[2].y = y;
#endif
        p[1].x = p[2].x + rx*c*sin(a1);
        p[1].y = p[2].y - ry*c*cos(a1);
        *last = p[2];
        transform_apply_pos(t, &p[0].x, &p[0].y);
        transform_apply_pos(t, &p[1].x, &p[1].y);
        transform_apply_pos(t, &p[2].x, &p[2].y);
    }
#endif
}


static void
_path_set_string(loa_buffer_t *vb, cairo_matrix_t *m, loa_path_t *path, const char *s)
{
    int op;
    loa_point_t start;
    loa_point_t last;
    path->segment = loa_buffer_segment_count(vb);
    path->segment_count = 0;

    op = 'M';
    start.x = 0;
    start.y = 0;
    last.x = 0;
    last.y = 0;
    for(;;){
        loa_point_t *p;
        SKIP_SPACE(s);
        if(!*s)
            return;
        if(isalpha(*s))
            op = *s++;
        switch(toupper(op)){
        case 'A':
            DEBUG_p("arc");
            //p = loa_buffer_push_segment(vb, path, LOA_LINE);
            {
                //point_t radius;
                loa_point_t end;
                double rx = _get_double(&s);
                double ry = _get_double(&s);
                double rot = _get_int(&s);
                double large = _get_int(&s);
                double sweep = _get_int(&s);
                _get_point(&end, &s, &last, op);
                DEBUG_p("rx: %f, ry: %f, rot: %f, large: %f, sweep: %f",rx,ry,rot,large,sweep);
                _loa_buffer_push_arc(vb, m, path, &last, rx, ry, rot, large, sweep, end.x, end.y);
            }
            break;
        case 'C':
            p = loa_buffer_push_segment(vb, path, LOA_CURVE);
            _get_point(&p[0], &s, &last, op);
            _get_point(&p[1], &s, &last, op);
            _get_point(&p[2], &s, &last, op);
            last = p[2];
            cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
            cairo_matrix_transform_point(m, &p[1].x, &p[1].y);
            cairo_matrix_transform_point(m, &p[2].x, &p[2].y);
            break;
        case 'H':
            //DEBUG_p("hline %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_LINE);
            p[0].x = _get_double(&s);
            if(islower(op))
                p[0].x += last.x;
            p[0].y = last.y;
            last.x = p[0].x;
            cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
            break;
        case 'L':
            //DEBUG_p("line %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_LINE);
            _get_point(&p[0], &s, &last, op);
            last = p[0];
            cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
            break;
        case 'M':
            //DEBUG_p("move %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_GOTO);
            _get_point(&p[0], &s, &last, op);
            start = last = p[0];
            cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
            op += 'L'-'M';
            break;
        case 'Q':
            p = loa_buffer_push_segment(vb, path, LOA_CURVE);
            _get_point(&p[0], &s, &last, op);
            _get_point(&p[2], &s, &last, op);
            last = p[2];
            cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
            cairo_matrix_transform_point(m, &p[2].x, &p[2].y);
            p[1] = p[0];
            break;
        case 'V':
            //DEBUG_p("vline %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_LINE);
            p[0].y = _get_double(&s);
            if(islower(op))
                p[0].y += last.y;
            p[0].x = last.x;
            last.y = p[0].y;
            cairo_matrix_transform_point(m, &p[0].x, &p[0].y);
            break;
        case 'Z':
            //DEBUG_p("vline %c",op);
            loa_buffer_push_segment(vb, path, LOA_CLOSE);
            last = start;
            break;
        default:
            DEBUG_p("UNKNOWN OP: %c",op);
            break;
        }
    }
}


static void
_path_copy(loa_buffer_t *vb, loa_path_t *dst, const loa_path_t *src, int deep)
{
    if(deep){
        /* FIXME!! */
    }else{
        *dst = *src;
    }
}


static void
_shape_copy(loa_buffer_t *vb, loa_object_t *dst, const loa_object_t *src, int deep)
{
    if(deep){
        /* FIXME!! */
    }else{
        *dst = *src;
    }
}


typedef enum{
    A_ATTRIBUTE,
    A_TRANSFORM,
    A_MOTION
}animation_type_t;

typedef enum{
    NT_NONE,
    NT_SHAPE,
    NT_GROUP,
    NT_DEFS
}node_type_t;

typedef struct{
    int refcount;
    node_type_t type;
    char *name;
    //int first_point;
    //int last_point;
    int first_shape;
    int last_shape;
    cairo_matrix_t matrix;
    loa_object_t group;
}node_t;


typedef struct{
    enum{
        S_ROOT,
        S_SVG,
        S_NODE,
        S_MOTION
    }state;
    abydos_plugin_handle_t *h;
    loa_buffer_t vb;
    hashmap_t node_map;
    loa_object_t *shape;                /* Current shape. */
    loa_anim_t *anim;                   /* Current animation. */
    svg_color_parser_t color_parser;
    int in_mask;
    loa_array_t node_stack;
    int defslevel;
    int blindlevel;
}parse_context_t;


static node_t *
_node_new(parse_context_t *c, const node_t *parent, const char *name)
{
    node_t *n = malloc(sizeof(node_t));
    if(name){
        n->refcount = 2;
        n->name = strdup(name);
        hashmap_add(&c->node_map, n->name, n);
    }else{
        n->refcount = 1;
        n->name = NULL;
    }
    n->group.first = loa_buffer_point_count(&c->vb);
    n->first_shape = loa_buffer_object_count(&c->vb);
    if(parent){
        n->matrix = parent->matrix;
    }else{
        cairo_matrix_init_identity(&n->matrix);
    }
    return n;
}


static void
_node_unref(node_t *n)
{
    if(--n->refcount <= 0){
        if(n->name)
            free(n->name);
        free(n);
    }
}


static node_t *
_get_top_node(parse_context_t *c)
{
    int len = loa_array_length(&c->node_stack);
    return len > 0 ? *(node_t**)loa_array_get(&c->node_stack, len-1) : NULL;
}


static node_type_t
_get_top_node_type(parse_context_t *c)
{
    node_t *n = _get_top_node(c);
    return n ? n->type : NT_NONE;
}


static node_t *
_push_node(parse_context_t *c, xml_attributes_t *attr, node_type_t type)
{
    const node_t *src = _get_top_node(c);
    node_t *dst = _node_new(c, src, xml_parser_attr_string(attr, "id"));
    node_t **pnode = loa_array_append(&c->node_stack, 1);
    *pnode = dst;

    dst->type = type;
    if(c->defslevel > 0){
        c->shape = &dst->group;
        //c->shape->style.fill_color.r = -1;
        //c->shape->style.stroke_color.r = -1;
    }else{
        c->shape = NT_SHAPE == type ? loa_buffer_add_object(&c->vb) : &dst->group;
        if(src){
            c->shape->style = src->group.style;
        }else{
            c->shape->style.fill_color.r = 0;
            c->shape->style.fill_color.g = 0;
            c->shape->style.fill_color.b = 0;
            c->shape->style.fill_color.a = 1;
            c->shape->style.stroke_color.r = -1;
            c->shape->style.stroke_color.a = 1;
            c->shape->style.stroke_width = 1;
        }
    }
    return dst;
}


static void
_pop_group(parse_context_t *c)
{
    node_t *n = _get_top_node(c);
    n->group.last = loa_buffer_point_count(&c->vb) - 1;
    n->last_shape = loa_buffer_object_count(&c->vb) - 1;
    _node_unref(n);
    loa_array_pop(&c->node_stack);
}


static node_t *
_get_xlink_node(parse_context_t *c, xml_attributes_t *attr)
{
    const char *xlink = xml_parser_attr_string(attr, "xlink:href");
    if(!xlink)
        return NULL;
    if('#' == *xlink)
        return hashmap_get(&(c)->node_map, xlink+1);
    return NULL;
}


static loa_object_t *
_get_xlink_shape(parse_context_t *c, xml_attributes_t *attr)
{
    node_t *n = _get_xlink_node(c, attr);
    return n ? &n->group : NULL;
}


typedef void(*adjust_func_t)(loa_keyframe_t*, node_t*);


static loa_keyframe_t *
_create_keyframe(parse_context_t *c, const char **value, const char **extra, int values, loa_anim_t *anim, adjust_func_t adjust_func, node_t *n)
{
    loa_keyframe_t *kf;
    int i;

    if(!**value)
        return NULL;

    kf = loa_buffer_add_keyframe(&c->vb);
    for(i=0;i<values;++i)
        kf->value[i] = _get_double(value);

    switch(anim->mode){
    case AM_DISCRETE:
    case AM_LINEAR:
    case AM_PACED:
        /* FIXME!! */
        break;
    case AM_SPLINE:
        kf->spline.c1.x = _get_double(extra);
        kf->spline.c1.y = _get_double(extra);
        kf->spline.c2.x = _get_double(extra);
        kf->spline.c2.y = _get_double(extra);
        break;
    }

    if(adjust_func)
        adjust_func(kf, n);

    return kf;
/*
error:
    --c->vb.keyframe_array.members;
    return NULL;
*/
}


static void
_adjust_rotate(loa_keyframe_t *kf, node_t *n)
{
    kf->rotate.a = kf->value[0] * M_PI / 180;
}


static void
_adjust_scale(loa_keyframe_t *kf, node_t *n)
{
    kf->scale.x = n->matrix.x0;
    kf->scale.y = n->matrix.y0;
}


static void
_adjust_ellipse_x(loa_keyframe_t *kf, node_t *n)
{
    kf->move.dx = kf->value[0];
    kf->move.dy = 0;
}


static void
_adjust_ellipse_y(loa_keyframe_t *kf, node_t *n)
{
    kf->move.dy = kf->value[0];
    kf->move.dx = 0;
}


static node_t *
_get_node(parse_context_t *c, xml_attributes_t *attr)
{
    node_t *n = _get_xlink_node(c, attr);
    return n ? n : _get_top_node(c);
}


static loa_object_t *
_get_shape(parse_context_t *c, xml_attributes_t *attr)
{
    loa_object_t *shape = _get_xlink_shape(c, attr);
    return shape ? shape : c->shape;
}


static int
_set_shape(loa_anim_t *anim, parse_context_t *c, xml_attributes_t *attr)
{
#if 1
    node_t *n = _get_node(c, attr);
    if(!n)
        return -1;
    anim->shape.first = n->first_shape;
    anim->shape.last = n->last_shape;
#else
    loa_object_t *shape = _get_shape(c, attr);
    if(!shape)
        return -1;
    anim->shape.first = anim->shape.last = loa_buffer_object_index(&c->vb, shape);
#endif
    return 0;
}


static int
_set_points(loa_anim_t *anim, parse_context_t *c, xml_attributes_t *attr)
{
#if 1
    loa_object_t *shape = _get_shape(c, attr);
    if(!shape){
        node_t *n = _get_top_node(c);
        if(!n)
            return -1;
        shape = &n->group;
    }
    anim->point.first = shape->first;
    anim->point.last = shape->last;
#else
    loa_object_t *shape = _get_shape(c, attr);
    if(shape){
        anim->point.first = shape->first;
        anim->point.last = shape->last;
    }else{
        node_t *node = _get_top_node(c);
        if(!node)
            return -1;
        anim->point.first = node->first_point;
        anim->point.last = node->last_point;
    }
#endif
    return 0;
}


static loa_anim_t *
_create_anim(parse_context_t *c, xml_attributes_t *attr, animation_type_t type)
{
    const char *s1;
    const char *s2;
    int values = 1;
    adjust_func_t adjust_func = NULL;
    c->anim = loa_buffer_add_anim(&c->vb);

    switch(type){
    case A_ATTRIBUTE:
        if(_set_shape(c->anim, c, attr) < 0)
            goto error;
        s1 = xml_parser_attr_string(attr, "attributeName");
        if(!s1)
            goto error;
        if(!strcasecmp(s1, "stroke-dasharray")){
            /* FIXME!! */
            values = 2;
        }else if(!strcasecmp(s1, "stroke-dashoffset")){
            /* FIXME!! */
        }else if(!strcasecmp(s1, "opacity")){
            c->anim->type = AT_FILL_ALPHA;
        }else{
            loa_object_t *shape = loa_buffer_get_object(&c->vb, c->anim->shape.first);
            switch(shape->type){
            case OT_GROUP:
                /* not reached */
                goto error;
            case OT_ELLIPSE:
                c->anim->point.first = shape->first;
                c->anim->point.last = c->shape->last;
                if(!strcasecmp(s1, "cx")){
                    c->anim->type = AT_MOVE;
                    adjust_func = _adjust_ellipse_x;
                }else if(!strcasecmp(s1, "cy")){
                    c->anim->type = AT_MOVE;
                    adjust_func = _adjust_ellipse_y;
                }else{
                    goto error;
                }
                break;
            case OT_PATH:
                goto error;
            }
        }
        break;
    case A_TRANSFORM:
        if(_set_points(c->anim, c, attr) < 0)
            goto error;
        s1 = xml_parser_attr_string(attr, "type");
        if(!s1)
            goto error;
        if(!strcasecmp(s1, "rotate")){
            c->anim->type = AT_ROTATE;
            values = 3;
            adjust_func = _adjust_rotate;
        }else if(!strcasecmp(s1, "scale")){
            c->anim->type = AT_SCALE;
            adjust_func = _adjust_scale;
        }else if(!strcasecmp(s1, "translate")){
            c->anim->type = AT_MOVE;
            values = 2;
        }else{
            goto error;
        }
        break;
    case A_MOTION:
        if(_set_points(c->anim, c, attr) < 0)
            goto error;
        c->anim->type = AT_MOTION;
        s1 = xml_parser_attr_string(attr, "path");
        if(s1){
            _path_set_string(&c->vb, &_get_top_node(c)->matrix, &c->anim->motion.path, s1);
        }else{
            c->anim->motion.path.segment_count = 0;
            /*
            shape_t *shape = _get_xlink_shape(c, attr);
            if(!shape)
                goto error;
            DEBUG_p("motion shape: %p", shape);
            _path_copy(&c->vb, &anim->motion.path, &shape->path, 0);
            */
        }
        //anim->motion.length = path_length(&anim->motion.path, loa_buffer_get_segment(&c->vb, 0), loa_buffer_get_point(&c->vb, 0));
        values = 0;
        break;
    }

    s1 = xml_parser_attr_string(attr, "dur");
    if(!s1)
        goto error;
    c->anim->duration = strtod(s1, NULL);
    //anim->begin = xml_parser_attr_double(attr, "begin", 0);

    if(values > 0){
        const char *extra;
        c->anim->keyframe.first = loa_buffer_keyframe_count(&c->vb);
        c->anim->mode = AM_LINEAR;
        s1 = xml_parser_attr_string(attr, "calcMode");
        if(s1){
            if(!strcmp(s1, "spline")){
                extra = xml_parser_attr_string(attr, "keySplines");
                if(extra)
                    c->anim->mode = AM_SPLINE;
            }
        }
        s1 = xml_parser_attr_string(attr, "keyTimes");
        s2 = xml_parser_attr_string(attr, "values");
        node_t *n = _get_node(c, attr);
        if(s1 && s2){
            for(;;){
                loa_keyframe_t *kf = _create_keyframe(c, &s2, &extra, values, c->anim, adjust_func, n);
                if(!kf)
                    break;
                kf->frac = _get_double(&s1);
            }
        }else{
            s1 = xml_parser_attr_string(attr, "from");
            s2 = xml_parser_attr_string(attr, "to");
            if(s1 && s2){
                loa_keyframe_t *kf = _create_keyframe(c, &s1, &extra, values, c->anim, adjust_func, n);
                if(!kf)
                    goto error;
                kf->frac = 0;
                kf = _create_keyframe(c, &s2, &extra, values, c->anim, adjust_func, n);
                if(!kf)
                    goto error;
                kf->frac = 1;
            }else{
                goto error;
            }
        }
        c->anim->keyframe.last = loa_buffer_keyframe_count(&c->vb) - 1;
    }

    return c->anim;
error:
    c->vb.keyframe_array.members = c->anim->keyframe.first;
    --c->vb.anim_array.members;
    return NULL;
}


static void
_on_element_begin(void *userdata, const char *e, xml_attributes_t *attr)
{
    parse_context_t *c = (parse_context_t*)userdata;
    DEBUG_p("begin: %s",e);
    if(c->blindlevel > 0){
        ++c->blindlevel;
        return;
    }
    switch(c->state){
    case S_ROOT:
        if(!strcasecmp(e, "svg")){
            const char *s = xml_parser_attr_string(attr, "viewBox");
            if(s){
                _get_int(&s);
                _get_int(&s);
                c->h->info->width = _get_int(&s);
                c->h->info->height = _get_int(&s);
            }else{
                c->h->info->width = xml_parser_attr_int(attr, "width", 0);
                c->h->info->height = xml_parser_attr_int(attr, "height", 0);
            }
            c->state = S_SVG;
            return;
        }
        break;
    case S_SVG:
    case S_NODE:
        if(_get_top_node_type(c) != NT_SHAPE){
            if(!strcasecmp(e, "defs")){
                ++c->defslevel;
                /* group_t *group = */ _push_node(c, attr, NT_DEFS);
                c->state = S_NODE;
                return;
            }else if(!strcasecmp(e, "g")){
                node_t *n = _push_node(c, attr, NT_GROUP);
                _style_set_attributes(&c->color_parser, &c->shape->style, attr);
                _matrix_multiply_string(&n->matrix, xml_parser_attr_string(attr, "transform"));
                c->state = S_NODE;
                return;
            }else if(!strcasecmp(e, "use")){
                node_t *n = _push_node(c, attr, NT_SHAPE);
                loa_object_t *sh = _get_xlink_shape(c, attr);
                if(sh){
                    _shape_copy(&c->vb, c->shape, sh, 0);
                    _style_set_attributes(&c->color_parser, &c->shape->style, attr);
                    _matrix_multiply_string(&n->matrix, xml_parser_attr_string(attr, "transform"));
                }
                c->state = S_NODE;
                return;
            }else if(!strcasecmp(e, "circle")){
                node_t *n = _push_node(c, attr, NT_SHAPE);
                loa_point_t *p = loa_buffer_add_points(&c->vb, 1);
                cairo_matrix_t *m = &n->matrix;
                c->shape->type = OT_ELLIPSE;
                c->shape->first = c->shape->last = loa_buffer_point_index(&c->vb, p);
                p->x = xml_parser_attr_int(attr, "cx", 0);
                p->y = xml_parser_attr_int(attr, "cy", 0);
                c->shape->ellipse.rw = c->shape->ellipse.rh = xml_parser_attr_int(attr, "r", 0);
                _style_set_attributes(&c->color_parser, &c->shape->style, attr);
                DEBUG_p("ellipse: %fx%f, %fx%f", p->x, p->y, c->shape->ellipse.rw, c->shape->ellipse.rh);
                cairo_matrix_transform_point(m, &p->x, &p->y);
                cairo_matrix_transform_distance(m, &c->shape->ellipse.rw, &c->shape->ellipse.rh);
                DEBUG_p("ellipse: %fx%f, %fx%f", p->x, p->y, c->shape->ellipse.rw, c->shape->ellipse.rh);
                c->state = S_NODE;
                return;
            }else if(!strcasecmp(e, "path")){
                node_t *n = _push_node(c, attr, NT_SHAPE);
                c->shape->type = OT_PATH;
                _matrix_multiply_string(&n->matrix, xml_parser_attr_string(attr, "transform"));
                _style_set_attributes(&c->color_parser, &c->shape->style, attr);
                DEBUG_p("fill: %f,%f,%f", c->shape->style.fill_color.r, c->shape->style.fill_color.g, c->shape->style.fill_color.b);
                _path_set_string(&c->vb, &n->matrix, &c->shape->path, xml_parser_attr_string(attr, "d"));
                c->state = S_NODE;
                return;
            }else if(!strcasecmp(e, "rect")){
                node_t *n = _push_node(c, attr, NT_SHAPE);
                loa_point_t *p;
                double x, y, w, h;
                c->shape->type = OT_PATH;
                _matrix_multiply_string(&n->matrix, xml_parser_attr_string(attr, "transform"));
                c->shape->first = loa_buffer_point_count(&c->vb);
                c->shape->path.segment = loa_buffer_segment_count(&c->vb);
                c->shape->path.point_count = 0;
                p = loa_buffer_push_segment(&c->vb, &c->shape->path, LOA_GOTO);
                p->x = x = xml_parser_attr_int(attr, "x", 0);
                p->y = y = xml_parser_attr_int(attr, "y", 0);
                cairo_matrix_transform_point(&n->matrix, &p->x, &p->y);
                w = xml_parser_attr_int(attr, "width", 0);
                h = xml_parser_attr_int(attr, "height", 0);
                //cairo_matrix_transform_distance(&group->matrix, &w, &h);
                p = loa_buffer_push_segment(&c->vb, &c->shape->path, LOA_LINE);
                p->x = x+w;
                p->y = y;
                cairo_matrix_transform_point(&n->matrix, &p->x, &p->y);
                p = loa_buffer_push_segment(&c->vb, &c->shape->path, LOA_LINE);
                p->x = x+w;
                p->y = y+h;
                cairo_matrix_transform_point(&n->matrix, &p->x, &p->y);
                p = loa_buffer_push_segment(&c->vb, &c->shape->path, LOA_LINE);
                p->x = x;
                p->y = y+h;
                cairo_matrix_transform_point(&n->matrix, &p->x, &p->y);
                p = loa_buffer_push_segment(&c->vb, &c->shape->path, LOA_CLOSE);
                c->shape->path.segment_count = loa_buffer_segment_count(&c->vb) - c->shape->path.segment;
                c->shape->last = c->shape->first + c->shape->path.point_count - 1;
                _style_set_attributes(&c->color_parser, &c->shape->style, attr);
#if 0
                DEBUG_p("rect: %fx%f, %fx%f", p->x, p->y, shape->rect.width, shape->rect.height);
                transform_apply_pos(&c->transform, &p->x, &p->y);
                transform_apply_size(&c->transform, &shape->rect.width, &shape->rect.height);
                DEBUG_p("rect: %fx%f, %fx%f", p->x, p->y, shape->rect.width, shape->rect.height);
#endif
                c->state = S_NODE;
                return;
            }
        }
        if(!strcasecmp(e, "animate")){
#if 1
            _create_anim(c, attr, A_ATTRIBUTE);
#else
            loa_anim_t *anim = loa_buffer_add_anim(&c->vb);
            const char* a = xml_parser_attr_string(attr, "attributeName");
            double from = xml_parser_attr_double(attr, "from", 0);
            double to = xml_parser_attr_double(attr, "to", 0);
            const char *values = xml_parser_attr_string(attr, "values");
            double d = to - from;
            if(!c->shape)
                break;
            if(!strcasecmp(a, "opacity")){
                anim->type = AT_FILL_ALPHA;
                anim->shape = loa_buffer_shape_index(&c->vb, c->shape);
                anim->alpha.value = to;
                anim->duration = xml_parser_attr_double(attr, "dur", 1);
            }else{
                switch(c->shape->type){
                case OT_TEMPLATE:
                    /* Not reached */
                    break;
                case OT_ELLIPSE:
                    if(!strcasecmp(a, "cx")){
                        anim->type = AT_MOVE;
                        anim->point.first = c->shape->first;
                        anim->point.last = c->shape->last;
                        anim->move.dx = d;
                        anim->move.dy = 0;
                        anim->duration = xml_parser_attr_double(attr, "dur", 1);
                    }else if(!strcasecmp(a, "cy")){
                        anim->type = AT_MOVE;
                        anim->point.first = c->shape->first;
                        anim->point.last = c->shape->last;
                        anim->move.dx = 0;
                        anim->move.dy = d;
                        anim->duration = xml_parser_attr_double(attr, "dur", 1);
                    }else{
                        --c->vb.anim_array.members;
                    }
                    break;
                case OT_PATH:
                    /* A path has no animatable attributes. */
                    break;
                }
            }
#endif
        }else if(!strcasecmp(e, "animateMotion")){
#if 1
            if(_create_anim(c, attr, A_MOTION)){
                c->state = S_MOTION;
                return;
            }
#else
            group_t *group = _get_group(c);
            anim_t *anim = loa_buffer_add_anim(&c->vb);
            const char *s;
            anim->type = AT_MOTION;
            anim->point.first = group->first_point;
            anim->point.last = /* c->shape->last = */ loa_buffer_point_count(&c->vb) - 1;
            s = xml_parser_attr_string(attr, "path");
            if(s){
                _path_set_string(&c->vb, &_get_group(c)->matrix, &anim->motion.path, xml_parser_attr_string(attr, "path"));
            }else{
                shape_t *shape = _get_xlink_shape(c, attr);
                DEBUG_p("motion shape: %p", shape);
                _path_copy(&c->vb, &anim->motion.path, &shape->path, 0);
            }
            anim->motion.length = path_length(&anim->motion.path, loa_buffer_get_segment(&c->vb, 0), loa_buffer_get_point(&c->vb, 0));
            anim->duration = xml_parser_attr_double(attr, "dur", 1);
#endif
        }else if(!strcasecmp(e, "animateTransform")){
            _create_anim(c, attr, A_TRANSFORM);
        }
        break;
    case S_MOTION:
        if(!strcasecmp(e, "mpath")){
            loa_object_t *shape = _get_xlink_shape(c, attr);
            if(shape){
                DEBUG_p("motion shape: %p", shape);
                _path_copy(&c->vb, &c->anim->motion.path, &shape->path, 0);
            }
        }
        break;
    }
    c->blindlevel = 1;
}


static void
_on_element_end(void *userdata, const char *e)
{
    parse_context_t *c = (parse_context_t*)userdata;
    DEBUG_p("end: %s",e);
    if(c->blindlevel > 0){
        --c->blindlevel;
        return;
    }
    switch(c->state){
    case S_ROOT:
        /* not reached */
        return;
    case S_SVG:
        c->state = S_ROOT;
        return;
    case S_NODE:
        if(NT_DEFS == _get_top_node_type(c))
            --c->defslevel;
        _pop_group(c);
        c->shape = NULL;
        if(loa_array_length(&c->node_stack) > 0){
            c->state = S_NODE;
        }else{
            c->state = S_SVG;
        }
        return;
    case S_MOTION:
        c->anim->motion.length = path_length(&c->anim->motion.path, loa_buffer_get_segment(&c->vb, 0), loa_buffer_get_point(&c->vb, 0));
        return;
    }
}


static int
_svg_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    xml_handle_t *xml;
    parse_context_t c;
    c.state = S_ROOT;
    c.h = h;
    loa_buffer_init(&c.vb, sizeof(loa_object_t), 0, sizeof(loa_anim_t), sizeof(loa_keyframe_t));
    //hashmap_init(&c.shape_map, str_hash, str_equal, free, NULL);
    hashmap_init(&c.node_map, str_hash, str_equal, NULL, (hashmap_notify_func_t)_node_unref);
    c.in_mask = 0;
    //transform_init(&c.transform);
    loa_array_init(&c.node_stack, sizeof(node_t*));
    svg_color_parser_init(&c.color_parser);
    //c.grouplevel = 0;
    c.defslevel = 0;
    c.blindlevel = 0;

    xml = xml_parser_begin(_on_element_begin, _on_element_end, &c);
    xml_parser_feed(xml, data, len);
    xml_parser_end(xml);

    h->point = loa_buffer_copy_points(&c.vb);
    h->point_count = loa_buffer_point_count(&c.vb);
    h->segment = loa_buffer_copy_segments(&c.vb);
    h->segment_count = loa_buffer_segment_count(&c.vb);
    h->shape = loa_buffer_copy_objects(&c.vb);
    h->shape_count = loa_buffer_object_count(&c.vb);
    h->anim = loa_buffer_copy_anims(&c.vb);
    h->anim_count = loa_buffer_anim_count(&c.vb);
    h->keyframe = loa_buffer_copy_keyframes(&c.vb);
    h->keyframe_count = loa_buffer_keyframe_count(&c.vb);

    loa_buffer_done(&c.vb);
    //hashmap_done(&c.shape_map);
    hashmap_done(&c.node_map);
    //transform_done(&c.transform);
    loa_array_done(&c.node_stack);
    svg_color_parser_done(&c.color_parser);
    return 0;
}


static double
_interpolate(double d1, double d2, double frac)
{
    return d1 + (d2 - d1) * frac;
}


static void
_svg_render(abydos_plugin_handle_t *h,
            cairo_t *cr,
            int page,
            cairo_bool_t *layer_visibility,
            int variant,
            double time,
            int frame)
{
    loa_point_t *point;
    loa_object_t *shape;
    int i;

    point = malloc(h->point_count * sizeof(loa_point_t));
    memcpy(point, h->point, h->point_count * sizeof(loa_point_t));

    shape = malloc(h->shape_count * sizeof(loa_object_t));
    memcpy(shape, h->shape, h->shape_count * sizeof(loa_object_t));

    for(i=0;i<h->anim_count;++i){
        const loa_anim_t *anim = &h->anim[i];
        const loa_keyframe_t *kf1;
        const loa_keyframe_t *kf2;
        double progress = fmod(time, anim->duration) / anim->duration;
        int j;

        if(anim->type != AT_MOTION){
            if(anim->keyframe.first >= anim->keyframe.last)
                continue;
            for(j=anim->keyframe.first+1;j<=anim->keyframe.last;++j){
                kf2 = &h->keyframe[j];
                if(kf2->frac >= progress)
                    break;
            }
            kf1 = &h->keyframe[j-1];
            progress = (progress - kf1->frac) / (kf2->frac - kf1->frac);
            switch(anim->mode){
            case AM_DISCRETE:
                progress = 0;
                break;
            case AM_LINEAR:
                break;
            case AM_PACED:
                /* FIXME!! */
                break;
            case AM_SPLINE:
                {
                    static const loa_point_t p1 = {0, 0};
                    static const loa_point_t p2 = {1, 1};
                    double x, y;
                    path_curve_trace(&p1, &kf1->spline.c1, &kf1->spline.c2, &p2, progress, &x, &y, NULL);
                    progress = y;
                }
                break;
            }
        }

        switch(anim->type){
        case AT_MOVE:
            {
                double dx = _interpolate(kf1->move.dx, kf2->move.dx, progress);
                double dy = _interpolate(kf1->move.dy, kf2->move.dy, progress);
                for(j=anim->point.first;j<=anim->point.last;++j){
                    loa_point_t *p = &point[j];
                    p->x += dx;
                    p->y += dy;
                }
            }
            break;
        case AT_MOTION:
            {
                cairo_matrix_t m;
                double x, y, a;
                cairo_matrix_init_identity(&m);
                //path_start(&anim->motion.path, h->segment, h->point, &x, &y);
                //cairo_matrix_translate(&m, -x, -y);
                path_trace(&anim->motion.path, anim->motion.length, h->segment, h->point, progress, &x, &y, &a);
                cairo_matrix_translate(&m, x, y);
                cairo_matrix_rotate(&m, a);
                for(j=anim->point.first;j<=anim->point.last;++j){
                    loa_point_t *p = &point[j];
                    cairo_matrix_transform_point(&m, &p->x, &p->y);
                }
            }
            break;
        case AT_ROTATE:
            {
                cairo_matrix_t m;
                double dx = _interpolate(kf1->rotate.x, kf2->rotate.x, progress);
                double dy = _interpolate(kf1->rotate.y, kf2->rotate.y, progress);
                cairo_matrix_init_identity(&m);
                cairo_matrix_translate(&m, dx, dy);
                cairo_matrix_rotate(&m, _interpolate(kf1->rotate.a, kf2->rotate.a, progress));
                cairo_matrix_translate(&m, -dx, -dy);
                for(j=anim->point.first;j<=anim->point.last;++j){
                    loa_point_t *p = &point[j];
                    cairo_matrix_transform_point(&m, &p->x, &p->y);
                }
            }
            break;
        case AT_SCALE:
            {
                cairo_matrix_t m;
                double scale = _interpolate(kf1->scale.s, kf2->scale.s, progress);
                cairo_matrix_init_identity(&m);
                cairo_matrix_translate(&m, kf1->scale.x, kf1->scale.y);
                cairo_matrix_scale(&m, scale, scale);
                cairo_matrix_translate(&m, -kf1->scale.x, -kf1->scale.y);
                for(j=anim->point.first;j<=anim->point.last;++j){
                    loa_point_t *p = &point[j];
                    cairo_matrix_transform_point(&m, &p->x, &p->y);
                }
            }
            break;
        case AT_FILL_ALPHA:
            {
                for(j=anim->shape.first;j<=anim->shape.last;++j){
                    loa_object_t *sh = &shape[j];
                    sh->style.fill_color.a = _interpolate(kf1->alpha, kf2->alpha, progress);
                }
            }
            break;
        }
    }

#if 0
    cairo_rectangle(cr, 0, 0, h->info->width, h->info->height);
    cairo_clip(cr);
#endif
    for(i=0;i<h->shape_count;++i){
        const loa_object_t *sh = &shape[i];
        const loa_point_t *p;
        int j;
        cairo_save(cr);
        switch(sh->type){
        case OT_GROUP:
            /* Not reached. */
            break;
        case OT_ELLIPSE:
            p = &point[sh->first];
            cairo_arc(cr, p->x, p->y, sh->ellipse.rw, 0, 2*M_PI);
            break;
        case OT_PATH:
            for(j=0;j<sh->path.segment_count;++j){
                const loa_segment_t *seg = &h->segment[sh->path.segment+j];
                loa_point_t *p = &point[seg->point];
                switch(seg->type){
                case LOA_GOTO:
                    cairo_move_to(cr, p[0].x, p[0].y);
                    break;
                case LOA_LINE:
                    cairo_line_to(cr, p[0].x, p[0].y);
                    break;
                case LOA_CURVE:
                    cairo_curve_to(cr, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y);
                    break;
                case LOA_CLOSE:
                    cairo_close_path(cr);
                    break;
                }
            }
            break;
        }
        if(sh->style.fill_color.r >= 0){
            cairo_set_source_rgba(cr, sh->style.fill_color.r, sh->style.fill_color.g, sh->style.fill_color.b, sh->style.fill_color.a);
            cairo_fill_preserve(cr);
        }
        if(sh->style.stroke_color.r >= 0){
            cairo_set_source_rgba(cr, sh->style.stroke_color.r, sh->style.stroke_color.g, sh->style.stroke_color.b, sh->style.stroke_color.a);
            cairo_set_line_width(cr, sh->style.stroke_width);
            cairo_stroke_preserve(cr);
        }
        cairo_new_path(cr);
        cairo_restore(cr);
    }
    free(point);
    free(shape);
}


static void
_svg_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    /* FIXME!! calculate total duration */
    duration[0] = h->anim[0].duration;
    for(i=1;i<h->anim_count;++i)
        if(h->anim[i].duration > duration[0])
            duration[0] = h->anim[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _svg_get_mime_types,
        .new = _svg_new,
        .free = _svg_free,
        .create_from_data = _svg_create_from_data,
        .render = _svg_render,
        .get_duration = _svg_get_duration
    };
    return &p;
}
